from collections import defaultdict
from io import BytesIO

from buidl.ecc import S256Point, Signature
from buidl.hd import HDPublicKey, ltrim_path
from buidl.helper import (
    base64_decode,
    base64_encode,
    child_to_path,
    encode_varstr,
    int_to_little_endian,
    little_endian_to_int,
    parse_binary_path,
    path_network,
    read_varint,
    read_varstr,
    serialize_key_value,
)
from buidl.op import op_code_to_number
from buidl.script import (
    RedeemScript,
    Script,
    WitnessScript,
)
from buidl.tx import Tx, TxOut
from buidl.witness import Witness


PSBT_MAGIC = b"\x70\x73\x62\x74"
PSBT_SEPARATOR = b"\xff"
PSBT_DELIMITER = b"\x00"
# PSBT global
PSBT_GLOBAL_UNSIGNED_TX = b"\x00"
PSBT_GLOBAL_XPUB = b"\x01"
# PSBT in
PSBT_IN_NON_WITNESS_UTXO = b"\x00"
PSBT_IN_WITNESS_UTXO = b"\x01"
PSBT_IN_PARTIAL_SIG = b"\x02"
PSBT_IN_SIGHASH_TYPE = b"\x03"
PSBT_IN_REDEEM_SCRIPT = b"\x04"
PSBT_IN_WITNESS_SCRIPT = b"\x05"
PSBT_IN_BIP32_DERIVATION = b"\x06"
PSBT_IN_FINAL_SCRIPTSIG = b"\x07"
PSBT_IN_FINAL_SCRIPTWITNESS = b"\x08"
PSBT_IN_POR_COMMITMENT = b"\x09"
# PSBT out
PSBT_OUT_REDEEM_SCRIPT = b"\x00"
PSBT_OUT_WITNESS_SCRIPT = b"\x01"
PSBT_OUT_BIP32_DERIVATION = b"\x02"


class MixedNetwork(Exception):
    pass


class SuspiciousTransaction(Exception):
    """
    Advanced/atypical transactions will raise a SuspiciousTransaction Exception, as these cannot be trivially summarized.
    This exception does not mean there is a problem with the transaction, just that it is too complex for simple summary.
    """

    pass


def path_to_child(path_component):
    if path_component[-1:] == "'":
        child_number = 0x80000000 + int(path_component[:-1])
    else:
        child_number = int(path_component)
    return child_number


def serialize_binary_path(path):
    path = path.lower().replace("h", "'")
    bin_path = b""
    for component in path.split("/")[1:]:
        bin_path += int_to_little_endian(path_to_child(component), 4)
    return bin_path


class NamedPublicKey(S256Point):
    def __repr__(self):
        return "Point:\n{}\nPath:\n{}:{}\n".format(
            self.sec().hex(), self.root_fingerprint.hex(), self.root_path
        )

    def add_raw_path_data(self, raw_path, network=None):
        self.root_fingerprint = raw_path[:4]
        self.root_path = parse_binary_path(raw_path[4:])
        self.raw_path = raw_path
        if network is None:
            self.network = path_network(self.root_path)
        else:
            self.network = network

    def replace_xfp(self, new_xfp):
        self.add_raw_path_data(
            raw_path=bytes.fromhex(new_xfp) + serialize_binary_path(self.root_path),
            network=self.network,
        )

    @classmethod
    def parse(cls, key, s, network=None):
        point = super().parse(key[1:])
        point.__class__ = cls
        point.add_raw_path_data(read_varstr(s), network=network)
        return point

    def serialize(self, prefix):
        return serialize_key_value(prefix + self.sec(), self.raw_path)


class NamedHDPublicKey(HDPublicKey):
    def __repr__(self):
        return "HD:\n{}\nPath:\n{}:{}\n".format(
            super().__repr__(), self.root_fingerprint.hex(), self.root_path
        )

    def add_raw_path_data(self, raw_path, network=None):
        self.root_fingerprint = raw_path[:4]
        bin_path = raw_path[4:]
        self.root_path = parse_binary_path(bin_path)
        if self.depth != len(bin_path) // 4:
            raise ValueError("raw path calculated depth and depth are different")
        if network is None:
            self.network = path_network(self.root_path)
        else:
            self.network = network
        self.raw_path = raw_path
        self.sync_point()

    def sync_point(self):
        self.point.__class__ = NamedPublicKey
        self.point.root_fingerprint = self.root_fingerprint
        self.point.root_path = self.root_path
        self.point.raw_path = self.raw_path
        self.point.network = self.network

    def child(self, index):
        child = super().child(index)
        child.__class__ = self.__class__
        child.root_fingerprint = self.root_fingerprint
        child.root_path = self.root_path + child_to_path(index)
        child.network = path_network(child.root_path)
        child.raw_path = self.raw_path + int_to_little_endian(index, 4)
        child.sync_point()
        return child

    def pubkey_lookup(self, max_child=9):
        lookup = {}
        for child_index in range(max_child + 1):
            child = self.child(child_index)
            lookup[child.sec()] = child
            lookup[child.hash160()] = child
        return lookup

    def redeem_script_lookup(self, max_external=9, max_internal=9):
        """Returns a dictionary of RedeemScripts associated with p2sh-p2wpkh for the BIP44 child ScriptPubKeys"""
        # create a lookup to send back
        lookup = {}
        # create the external child (0)
        external = self.child(0)
        # loop through to the maximum external child + 1
        for child_index in range(max_external + 1):
            # grab the child at the index
            child = external.child(child_index)
            # create the p2sh-p2wpkh RedeemScript of [0, hash160]
            redeem_script = RedeemScript([0, child.hash160()])
            # hash160 of the RedeemScript is the key, RedeemScript is the value
            lookup[redeem_script.hash160()] = redeem_script
        # create the internal child (1)
        internal = self.child(1)
        # loop through to the maximum internal child + 1
        for child_index in range(max_internal + 1):
            # grab the child at the index
            child = internal.child(child_index)
            # create the p2sh-p2wpkh RedeemScript of [0, hash160]
            redeem_script = RedeemScript([0, child.hash160()])
            # hash160 of the RedeemScript is the key, RedeemScript is the value
            lookup[redeem_script.hash160()] = redeem_script
        # return the lookup
        return lookup

    def bip44_lookup(self, max_external=9, max_internal=9):
        external = self.child(0)
        internal = self.child(1)
        return {
            **external.pubkey_lookup(max_external),
            **internal.pubkey_lookup(max_internal),
        }

    @classmethod
    def parse(cls, key, s, network=None):
        hd_key = cls.raw_parse(BytesIO(key[1:]))
        hd_key.__class__ = cls
        hd_key.add_raw_path_data(read_varstr(s), network=network)
        return hd_key

    @classmethod
    def from_hd_pub(cls, child_hd_pub, xfp_hex, path):
        hd_key = child_hd_pub
        hd_key.__class__ = cls
        hd_key.add_raw_path_data(
            raw_path=bytes.fromhex(xfp_hex) + serialize_binary_path(path),
            network=hd_key.network,
        )
        return hd_key

    @classmethod
    def from_hd_priv(cls, root_hd_priv, path):
        child_hd_pub = root_hd_priv.traverse(path).pub
        return cls.from_hd_pub(
            child_hd_pub=child_hd_pub,
            xfp_hex=root_hd_priv.fingerprint().hex(),
            path=path,
        )

    def serialize(self):
        return serialize_key_value(
            PSBT_GLOBAL_XPUB + self.raw_serialize(), self.raw_path
        )

    def is_ancestor(self, named_pubkey):
        return named_pubkey.raw_path.startswith(self.raw_path)

    def verify_descendent(self, named_pubkey):
        if not self.is_ancestor(named_pubkey):
            raise ValueError("path is not a descendent of this key")
        remainder = named_pubkey.raw_path[len(self.raw_path) :]
        current = self
        while len(remainder):
            child_index = little_endian_to_int(remainder[:4])
            current = current.child(child_index)
            remainder = remainder[4:]
        return current.point == named_pubkey


class PSBT:
    def __init__(
        self,
        tx_obj,
        psbt_ins,
        psbt_outs,
        hd_pubs=None,
        extra_map=None,
        network="mainnet",
    ):
        self.tx_obj = tx_obj
        self.psbt_ins = psbt_ins
        self.psbt_outs = psbt_outs
        self.hd_pubs = hd_pubs or {}
        self.extra_map = extra_map or {}
        self.network = network
        self.tx_obj.network = network
        self.validate()

    def validate(self):
        """Checks the PSBT for consistency"""
        if len(self.tx_obj.tx_ins) != len(self.psbt_ins):
            raise ValueError(
                "Number of psbt_ins in the transaction should match the psbt_ins array"
            )
        for i, psbt_in in enumerate(self.psbt_ins):
            # validate the input
            psbt_in.validate()
            tx_in = self.tx_obj.tx_ins[i]
            if tx_in.script_sig.commands:
                raise ValueError("ScriptSig for the tx should not be defined")
            # validate the ScriptSig
            if psbt_in.script_sig:
                tx_in.script_sig = psbt_in.script_sig
                tx_in.witness = psbt_in.witness
                if not self.tx_obj.verify_input(i):
                    raise ValueError(
                        "ScriptSig/Witness at input {} provided, but not valid".format(
                            i
                        )
                    )
                tx_in.script_sig = Script()
                tx_in.witness = Witness()
            # validate the signatures
            if psbt_in.sigs:
                for sec, sig in psbt_in.sigs.items():
                    point = S256Point.parse(sec)
                    signature = Signature.parse(sig[:-1])
                    if psbt_in.prev_out:
                        # segwit
                        if not self.tx_obj.check_sig_segwit(
                            i,
                            point,
                            signature,
                            psbt_in.redeem_script,
                            psbt_in.witness_script,
                        ):
                            raise ValueError(
                                "segwit signature provided does not validate"
                            )
                    elif psbt_in.prev_tx:
                        # legacy
                        if not self.tx_obj.check_sig_legacy(
                            i, point, signature, psbt_in.redeem_script
                        ):
                            raise ValueError(
                                "legacy signature provided does not validate {}".format(
                                    self
                                )
                            )
            # validate the NamedPublicKeys
            if psbt_in.named_pubs:
                for named_pub in psbt_in.named_pubs.values():
                    named_pub.network = self.network
                    for hd_pub in self.hd_pubs.values():
                        if hd_pub.is_ancestor(named_pub):
                            if not hd_pub.verify_descendent(named_pub):
                                raise ValueError(
                                    "public key {} does not derive from xpub {}".format(
                                        named_pub, hd_pub
                                    )
                                )
                            break
        if len(self.tx_obj.tx_outs) != len(self.psbt_outs):
            raise ValueError(
                "Number of psbt_outs in the transaction should match the psbt_outs array"
            )
        for psbt_out in self.psbt_outs:
            # validate output
            psbt_out.validate()
            # validate the NamedPublicKeys
            if psbt_out.named_pubs:
                for named_pub in psbt_out.named_pubs.values():
                    named_pub.network = self.network
                    for hd_pub in self.hd_pubs.values():
                        if hd_pub.is_ancestor(named_pub):
                            if not hd_pub.verify_descendent(named_pub):
                                raise ValueError(
                                    "public key {} does not derive from xpub {}".format(
                                        named_pub, hd_pub
                                    )
                                )
                            break
        return True

    def __repr__(self):
        return "Tx:\n{}\nPSBT XPUBS:\n{}\nPsbt_Ins:\n{}\nPsbt_Outs:\n{}\nExtra:{}\n".format(
            self.tx_obj, self.hd_pubs, self.psbt_ins, self.psbt_outs, self.extra_map
        )

    @classmethod
    def create(
        cls,
        tx_obj,
        validate=True,
        tx_lookup={},
        pubkey_lookup={},
        redeem_lookup={},
        witness_lookup={},
        hd_pubs={},
    ):
        """Create a PSBT from a transaction"""

        # create an array of PSBTIns
        psbt_ins = []
        # iterate through the inputs of the transaction
        for tx_in in tx_obj.tx_ins:
            # Empty ScriptSig and Witness fields
            # if ScriptSig exists, save it then empty it
            if tx_in.script_sig.commands:
                script_sig = tx_in.script_sig
                tx_in.script_sig = Script()
            else:
                script_sig = None
            # if Witness exists, save it then empty it
            if tx_in.witness:
                witness = tx_in.witness
                tx_in.witness = Witness()
            else:
                witness = None
            # Create a PSBTIn with the TxIn, ScriptSig and Witness
            psbt_in = PSBTIn(tx_in, script_sig=script_sig, witness=witness)
            # add PSBTIn to array
            psbt_ins.append(psbt_in)

        # create an array of PSBTOuts
        psbt_outs = []
        # iterate through the outputs of the transaction
        for tx_out in tx_obj.tx_outs:
            # create the PSBTOut with the TxOut
            psbt_out = PSBTOut(tx_out)
            # add PSBTOut to arary
            psbt_outs.append(psbt_out)
        # return an instance with the Tx, PSBTIn array and PSBTOut array
        psbt_obj = cls(
            tx_obj=tx_obj,
            psbt_ins=psbt_ins,
            psbt_outs=psbt_outs,
            hd_pubs=None,
            extra_map=None,
            network=tx_obj.network,
        )

        # append metadata
        if tx_lookup or pubkey_lookup:
            psbt_obj.update(
                tx_lookup=tx_lookup,
                pubkey_lookup=pubkey_lookup,
                redeem_lookup=redeem_lookup,
                witness_lookup=witness_lookup,
            )

        if hd_pubs:
            psbt_obj.hd_pubs = hd_pubs

        if validate:
            psbt_obj.validate()

        return psbt_obj

    def update(self, tx_lookup, pubkey_lookup, redeem_lookup={}, witness_lookup={}):
        # update each PSBTIn
        for psbt_in in self.psbt_ins:
            psbt_in.update(
                tx_lookup=tx_lookup,
                pubkey_lookup=pubkey_lookup,
                redeem_lookup=redeem_lookup,
                witness_lookup=witness_lookup,
            )
        # update each PSBTOut
        for psbt_out in self.psbt_outs:
            psbt_out.update(
                pubkey_lookup=pubkey_lookup,
                redeem_lookup=redeem_lookup,
                witness_lookup=witness_lookup,
            )

    def sign(self, hd_priv):
        """Signs appropriate inputs with the hd private key provided"""
        # set the signed boolean to False until we sign something
        signed = False
        # grab the fingerprint of the private key
        fingerprint = hd_priv.fingerprint()
        # iterate through each PSBTIn
        for i, psbt_in in enumerate(self.psbt_ins):
            # iterate through the public keys associated with the PSBTIn
            for named_pub in psbt_in.named_pubs.values():
                # if the fingerprints match
                if named_pub.root_fingerprint == fingerprint:
                    # get the private key at the root_path of the NamedPublicKey
                    private_key = hd_priv.traverse(named_pub.root_path).private_key
                    if psbt_in.use_segwit_signature():
                        sig = self.tx_obj.get_sig_segwit(
                            i,
                            private_key,
                            psbt_in.redeem_script,
                            psbt_in.witness_script,
                        )
                    else:
                        sig = self.tx_obj.get_sig_legacy(
                            i, private_key, psbt_in.redeem_script
                        )
                    # update the sigs dict of the PSBTIn object
                    psbt_in.sigs[private_key.point.sec()] = sig
                    signed = True
        return signed

    def sign_with_private_keys(self, private_keys):
        """Signs appropriate inputs with the private key provided"""
        # set the signed boolean to False until we sign something
        signed = False
        # iterate through each private key
        for private_key in private_keys:
            # grab the point associated with the point
            point = private_key.point
            # iterate through each PSBTIn
            for i, psbt_in in enumerate(self.psbt_ins):
                # if the sec is in the named_pubs dictionary
                if psbt_in.named_pubs.get(point.sec()):
                    if psbt_in.use_segwit_signature():
                        sig = self.tx_obj.get_sig_segwit(
                            i,
                            private_key,
                            psbt_in.redeem_script,
                            psbt_in.witness_script,
                        )
                    else:
                        sig = self.tx_obj.get_sig_legacy(
                            i, private_key, psbt_in.redeem_script
                        )
                    # update the sigs dict of the PSBTIn object
                    psbt_in.sigs[private_key.point.sec()] = sig
                    signed = True
        # return whether we signed something
        return signed

    def combine(self, other):
        """combines information from another PSBT to this one"""
        # the tx_obj properties should be the same or raise a ValueError
        if self.tx_obj.hash() != other.tx_obj.hash():
            raise ValueError(
                "cannot combine PSBTs that refer to different transactions"
            )
        # combine the hd_pubs
        self.hd_pubs = {**other.hd_pubs, **self.hd_pubs}
        # combine extra_map
        self.extra_map = {**other.extra_map, **self.extra_map}
        # combine psbt_ins
        for psbt_in_1, psbt_in_2 in zip(self.psbt_ins, other.psbt_ins):
            psbt_in_1.combine(psbt_in_2)
        # combine psbt_outs
        for psbt_out_1, psbt_out_2 in zip(self.psbt_outs, other.psbt_outs):
            psbt_out_1.combine(psbt_out_2)

    def finalize(self):
        """Finalize the transaction by filling in the ScriptSig and Witness fields for each input"""
        # iterate through the inputs
        for psbt_in in self.psbt_ins:
            # finalize each input
            psbt_in.finalize()

    def final_tx(self):
        """Returns the broadcast-able transaction"""
        # clone the transaction from self.tx_obj
        tx_obj = self.tx_obj.clone()
        # determine if the transaction is segwit by looking for a witness field
        #  in any PSBTIn. if so, set tx_obj.segwit = True
        if any([psbt_in.witness for psbt_in in self.psbt_ins]):
            tx_obj.segwit = True
        # iterate through the transaction and PSBT inputs together
        #  using zip(tx_obj.tx_ins, self.psbt_ins)
        for tx_in, psbt_in in zip(tx_obj.tx_ins, self.psbt_ins):
            # set the ScriptSig of the transaction input
            tx_in.script_sig = psbt_in.script_sig
            # Exercise 7: if the tx is segwit, set the witness as well
            if tx_obj.segwit:
                # witness should be the PSBTIn witness or an empty Witness()
                tx_in.witness = psbt_in.witness or Witness()
        # check to see that the transaction verifies
        if not tx_obj.verify():
            raise RuntimeError("transaction invalid")
        # return the now filled in transaction
        return tx_obj

    @classmethod
    def parse_base64(cls, b64, network=None):
        stream = BytesIO(base64_decode(b64))
        return cls.parse(stream, network=network)

    @classmethod
    def parse(cls, s, network=None):
        """Returns an instance of PSBT from a stream"""
        # prefix
        magic = s.read(4)
        if magic != PSBT_MAGIC:
            raise SyntaxError("Incorrect magic")
        separator = s.read(1)
        if separator != PSBT_SEPARATOR:
            raise SyntaxError("No separator")
        # global data
        tx_obj = None
        hd_pubs = {}
        extra_map = {}
        key = read_varstr(s)
        while key != b"":
            psbt_type = key[0:1]
            if psbt_type == PSBT_GLOBAL_UNSIGNED_TX:
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if tx_obj:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                _ = read_varint(s)
                tx_obj = Tx.parse_legacy(s)
            elif psbt_type == PSBT_GLOBAL_XPUB:
                if len(key) != 79:
                    raise KeyError("Wrong length for the key")
                hd_pub = NamedHDPublicKey.parse(key, s, network=network)
                hd_pubs[hd_pub.raw_serialize()] = hd_pub
                if network is None:
                    network = hd_pub.network
                if hd_pub.network != network:
                    raise MixedNetwork("PSBT Mainnet/Testnet Mixing")
            else:
                if extra_map.get(key):
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                extra_map[key] = read_varstr(s)
            key = read_varstr(s)
        if not tx_obj:
            raise SyntaxError("transaction is required")
        # per input data
        psbt_ins = []
        for tx_in in tx_obj.tx_ins:
            psbt_in = PSBTIn.parse(s, tx_in, network=network)
            for named_pub in psbt_in.named_pubs.values():
                if network is None:
                    network = named_pub.network
                if named_pub.network != network:
                    raise MixedNetwork("PSBTIn Mainnet/Testnet Mixing")
            psbt_ins.append(psbt_in)
        # per output data
        psbt_outs = []
        for tx_out in tx_obj.tx_outs:
            psbt_out = PSBTOut.parse(s, tx_out, network=network)
            for named_pub in psbt_out.named_pubs.values():
                if network is None:
                    network = named_pub.network
                if named_pub.network != network:
                    raise MixedNetwork("PSBTOut Mainnet/Testnet Mixing")
            psbt_outs.append(psbt_out)
        return cls(tx_obj, psbt_ins, psbt_outs, hd_pubs, extra_map, network)

    def serialize_base64(self):
        return base64_encode(self.serialize())

    def serialize(self):
        # always start with the magic and separator
        result = PSBT_MAGIC + PSBT_SEPARATOR
        # tx
        result += serialize_key_value(PSBT_GLOBAL_UNSIGNED_TX, self.tx_obj.serialize())
        # xpubs
        for xpub in sorted(self.hd_pubs.keys()):
            hd_pub = self.hd_pubs[xpub]
            result += hd_pub.serialize()
        for key in sorted(self.extra_map.keys()):
            result += serialize_key_value(key, self.extra_map[key])
        # delimiter
        result += PSBT_DELIMITER
        # per input data
        for psbt_in in self.psbt_ins:
            result += psbt_in.serialize()
        # per output data
        for psbt_out in self.psbt_outs:
            result += psbt_out.serialize()
        return result

    def replace_root_xfps(self, xfp_map):
        """
        Can be used for extra privacy to blind an XFP not relevant to that signer.

        Supply an xfp_map of the following form:
        {
            # to_hide: to_replace
            "deadbeef": "00000000",
            "ab01ab01": "ffffffff",
        }
        """
        for xfp_to_hide, xfp_to_replace in xfp_map.items():
            was_replaced = False

            # replace from inputs
            for psbt_in in self.psbt_ins:
                for named_pub in psbt_in.named_pubs.values():
                    if named_pub.root_fingerprint.hex() == xfp_to_hide:
                        named_pub.replace_xfp(new_xfp=xfp_to_replace)
                        was_replaced = True

            # replace from outputs
            for psbt_out in self.psbt_outs:
                for named_pub in psbt_out.named_pubs.values():
                    if named_pub.root_fingerprint.hex() == xfp_to_hide:
                        named_pub.replace_xfp(new_xfp=xfp_to_replace)
                        was_replaced = True

            if not was_replaced:
                raise ValueError(f"xfp_hex {xfp_to_hide} not found in psbt")

    def _describe_basic_multisig_inputs(self, hdpubkey_map):

        # These will be used for all inputs and change outputs
        inputs_quorum_m, inputs_quorum_n = None, None

        # Gather TX info and validate
        inputs_desc = []
        total_input_sats = 0

        root_paths_for_signing = defaultdict(set)
        for cnt, psbt_in in enumerate(self.psbt_ins):
            psbt_in.validate()

            if psbt_in.witness_script and psbt_in.redeem_script:
                raise SuspiciousTransaction(
                    f"multisig input #{cnt} has both witness_script and redeem_script: {psbt_in}"
                )

            script_for_psbtin = psbt_in.witness_script or psbt_in.redeem_script

            if not script_for_psbtin:
                raise SuspiciousTransaction(
                    f"multisig input #{cnt} does not contain a script to evaluate"
                )

            # Be sure all xpubs are properly accounted for
            if len(hdpubkey_map) != len(psbt_in.named_pubs):
                # TODO: doesn't handle case where the same xfp is >1 signers
                raise SuspiciousTransaction(
                    f"{len(hdpubkey_map)} xpubs supplied != {len(psbt_in.named_pubs)} named_pubs in PSBT input."
                )

            input_quorum_m, input_quorum_n = script_for_psbtin.get_quorum()
            if inputs_quorum_m is None:
                inputs_quorum_m = input_quorum_m
            else:
                if inputs_quorum_m != input_quorum_m:
                    raise SuspiciousTransaction(
                        f"Previous input(s) set a quorum threshold of {inputs_quorum_m}, but this input is {input_quorum_m}"
                    )

            if inputs_quorum_n is None:
                inputs_quorum_n = input_quorum_n
                if inputs_quorum_n != len(hdpubkey_map):
                    raise SuspiciousTransaction(
                        f"Transaction has {len(hdpubkey_map)} pubkeys but we are expecting {input_quorum_n}"
                    )
            else:
                if inputs_quorum_n != input_quorum_n:
                    raise SuspiciousTransaction(
                        f"Previous input(s) set a max quorum of threshold of {inputs_quorum_n}, but this input is {input_quorum_n}"
                    )

            spend_addr = script_for_psbtin.address(network=self.network)

            bip32_derivs = []
            for named_pub in psbt_in.named_pubs.values():
                # Match to corresponding xpub to validate that this xpub is a participant in this input
                xfp = named_pub.root_fingerprint.hex()

                try:
                    hdpub = hdpubkey_map[xfp]
                except KeyError:
                    raise SuspiciousTransaction(
                        f"Root fingerprint {xfp} for input #{cnt} not in the hdpubkey_map you supplied"
                    )

                trimmed_path = ltrim_path(named_pub.root_path, depth=hdpub.depth)
                if hdpub.traverse(trimmed_path).sec() != named_pub.sec():
                    raise SuspiciousTransaction(
                        f"xpub {hdpub} with path {named_pub.root_path} does not appear to be part of input # {cnt}"
                    )

                root_paths_for_signing[xfp].add(named_pub.root_path)

                # this is very similar to what bitcoin-core's decodepsbt returns
                bip32_derivs.append(
                    {
                        "pubkey": named_pub.sec().hex(),
                        "master_fingerprint": xfp,
                        "path": named_pub.root_path,
                        "xpub": hdpub.xpub(),
                    }
                )

            # BIP67 sort order
            bip32_derivs = sorted(bip32_derivs, key=lambda k: k["pubkey"])

            input_sats = psbt_in.tx_in.value()
            total_input_sats += input_sats

            input_desc = {
                "quorum": f"{input_quorum_m}-of-{input_quorum_n}",
                "bip32_derivs": bip32_derivs,
                "prev_txhash": psbt_in.tx_in.prev_tx.hex(),
                "prev_idx": psbt_in.tx_in.prev_index,
                "n_sequence": psbt_in.tx_in.sequence,
                "sats": input_sats,
                "addr": spend_addr,
                "witness_script": str(psbt_in.witness_script)
                if psbt_in.witness_script
                else None,
                "redeem_script": str(psbt_in.redeem_script)
                if psbt_in.redeem_script
                else None,
            }
            inputs_desc.append(input_desc)

        if not root_paths_for_signing:
            raise SuspiciousTransaction(
                "No `root_paths_for_signing` with `hdpubkey_map` {hdpubkey_map} in PSBT:\n{self}"
            )

        return {
            "inputs_quorum_m": inputs_quorum_m,
            "inputs_quorum_n": inputs_quorum_n,
            "inputs_desc": inputs_desc,
            "root_paths_for_signing": root_paths_for_signing,
            "total_input_sats": total_input_sats,
        }

    def _describe_basic_multisig_outputs(
        self,
        expected_quorum_m,
        expected_quorum_n,
        hdpubkey_map={},
    ):

        # intialize variable we'll loop through to set
        outputs_desc = []
        spend_addr, spend_sats = "", 0
        change_addr, change_sats = "", 0
        total_sats = 0
        spends_cnt = 0
        for cnt, psbt_out in enumerate(self.psbt_outs):
            psbt_out.validate()

            output_desc = {
                "sats": psbt_out.tx_out.amount,
                "addr": psbt_out.tx_out.script_pubkey.address(network=self.network),
                "addr_type": psbt_out.tx_out.script_pubkey.__class__.__name__.rstrip(
                    "ScriptPubKey"
                ),
            }
            total_sats += output_desc["sats"]

            if psbt_out.named_pubs:
                # Output claims to be change, we will confirm it below before accepting that as fact

                script_for_psbtout = psbt_out.witness_script or psbt_out.redeem_script

                # FIXME: Confirm this works with a fake change test case
                output_quorum_m, output_quorum_n = script_for_psbtout.get_quorum()
                if expected_quorum_m != output_quorum_m:
                    raise SuspiciousTransaction(
                        f"Previous output(s) set a max quorum of threshold of {expected_quorum_m}, but this transaction is {output_quorum_m}"
                    )
                if expected_quorum_n != output_quorum_n:
                    raise SuspiciousTransaction(
                        f"Previous input(s) set a max cosigners of {expected_quorum_n}, but this transaction is {output_quorum_n}"
                    )

                # Be sure all xpubs are properly acocunted for
                if output_quorum_n != len(psbt_out.named_pubs):
                    # TODO: doesn't handle case where the same xfp is >1 signers (surprisngly complex)
                    raise SuspiciousTransaction(
                        f"{len(hdpubkey_map)} xpubs supplied != {len(psbt_out.named_pubs)} named_pubs in PSBT change output."
                        "You may be able to get this wallet to cosign a sweep transaction (1-output) instead."
                    )

                bip32_derivs = []
                for named_pub in psbt_out.named_pubs.values():
                    # Match to corresponding xpub to validate that this xpub is a participant in this change output
                    xfp = named_pub.root_fingerprint.hex()

                    try:
                        hdpub = hdpubkey_map[xfp]
                    except KeyError:
                        raise SuspiciousTransaction(
                            f"Root fingerprint {xfp} for output #{cnt} not in the hdpubkey_map you supplied"
                            "Do a sweep transaction (1-output) if you want this wallet to cosign."
                        )

                    trimmed_path = ltrim_path(named_pub.root_path, depth=hdpub.depth)
                    if hdpub.traverse(trimmed_path).sec() != named_pub.sec():
                        raise SuspiciousTransaction(
                            f"xpub {hdpub} with path {named_pub.root_path} does not appear to be part of output # {cnt}"
                            "You may be able to get this wallet to cosign a sweep transaction (1-output) instead."
                        )

                    # this is very similar to what bitcoin-core's decodepsbt returns
                    bip32_derivs.append(
                        {
                            "pubkey": named_pub.sec().hex(),
                            "master_fingerprint": xfp,
                            "path": named_pub.root_path,
                            "xpub": hdpub.xpub(),
                        }
                    )

                # BIP67 sort order
                bip32_derivs = sorted(bip32_derivs, key=lambda k: k["pubkey"])

                # Confirm there aren't >1 change ouputs
                # (this is technically allowed but too sketchy to support)
                if change_sats or change_addr:
                    raise SuspiciousTransaction(
                        f"Cannot have >1 change output.\n{outputs_desc}"
                    )
                change_addr = output_desc["addr"]
                change_sats = output_desc["sats"]

                output_desc["witness_script"] = (
                    str(psbt_out.witness_script) if psbt_out.witness_script else None
                )
                output_desc["redeem_script"] = (
                    str(psbt_out.redeem_script) if psbt_out.redeem_script else None
                )

                # We have succesfully validated that this output matches the input parameters:
                output_desc["is_change"] = True

            else:
                output_desc["is_change"] = False
                spends_cnt += 1
                spend_sats += output_desc["sats"]

                if spends_cnt > 1:
                    # There is no concept of a single spend addr/amount for batch spends
                    # Caller must interpret the batch spend from the returned outputs_desc
                    spend_addr = ""
                else:
                    spend_addr = output_desc["addr"]

            outputs_desc.append(output_desc)

        return {
            "total_sats": total_sats,
            "outputs_desc": outputs_desc,
            "change_addr": change_addr,
            "change_sats": change_sats,
            "spend_addr": spend_addr,
            "spend_sats": spend_sats,
            "is_batch_tx": spends_cnt > 1,
        }

    def describe_basic_multisig(self, hdpubkey_map={}):
        """
        Describe a typical multisig transaction in a human-readable way for manual verification before signing.

        This tool supports transactions with the following constraints:
        * All inputs have the exact same multisig wallet (quorum + xpubs)
        * All outputs are either spend or proven to be change. For UX reasons, there can not be >1 change address.

        A SuspiciousTransaction Exception does not strictly mean there is a problem with the transaction, it is likely just too complex for simple summary.

        Due to the nature of how PSBT works, if your PSBT is slimmed down (doesn't contain xpubs AND prev TX hexes), you must supply a `hdpubkey_map` for ALL n xpubs:
          {
            'xfphex1': HDPublicKey1,
            'xfphex2': HDPublicKey2,
          }
        These HDPublicKey's will be traversed according to the paths given in the PSBT.

        TODOS:
          - add helper method that accepts an output descriptor (converts it into an hdpubkey_map and then calls this method)
          - add support for p2wsh-wrapped-p2sh?
        """

        self.validate()

        tx_fee_sats = self.tx_obj.fee()

        if not hdpubkey_map:
            if not self.hd_pubs:
                raise ValueError(
                    "Cannot describe multisig PSBT without `hd_pubs` nor `hdpubkey_map`"
                )
            # build hdpubkey_map from PSBT's hdpubs
            hdpubkey_map = {}
            for hdpubkey in self.hd_pubs.values():
                hdpubkey_map[hdpubkey.root_fingerprint.hex()] = HDPublicKey.parse(
                    hdpubkey.xpub()
                )

        inputs_described = self._describe_basic_multisig_inputs(
            hdpubkey_map=hdpubkey_map,
        )
        total_input_sats = inputs_described["total_input_sats"]

        outputs_described = self._describe_basic_multisig_outputs(
            hdpubkey_map=hdpubkey_map,
            # Tool requires m-of-n be same for inputs as outputs
            expected_quorum_m=inputs_described["inputs_quorum_m"],
            expected_quorum_n=inputs_described["inputs_quorum_n"],
        )
        is_batch_tx = outputs_described["is_batch_tx"]
        total_output_sats = outputs_described["total_sats"]
        spend_sats = outputs_described["spend_sats"]
        spend_addr = outputs_described["spend_addr"]

        tx_fee_rounded = round(tx_fee_sats / total_input_sats * 100, 2)
        # comma separating satoshis for better display
        if is_batch_tx:
            spend_breakdown = "\n".join(
                [
                    f"{x['addr']}: {x['sats']:,} sats"
                    for x in outputs_described["outputs_desc"]
                    if not x["is_change"]
                ]
            )
            tx_summary_text = f"Batch PSBT sends {spend_sats:,} sats with a fee of {tx_fee_sats:,} sats ({tx_fee_rounded}% of spend). Batch spend breakdown:\n{spend_breakdown}"
        else:
            tx_summary_text = f"PSBT sends {spend_sats:,} sats to {spend_addr} with a fee of {tx_fee_sats:,} sats ({tx_fee_rounded}% of spend)"

        return {
            # TX level:
            "txid": self.tx_obj.id(),
            "tx_summary_text": tx_summary_text,
            "locktime": self.tx_obj.locktime,
            "version": self.tx_obj.version,
            "network": self.network,
            "tx_fee_sats": tx_fee_sats,
            "total_input_sats": total_input_sats,
            "total_output_sats": total_output_sats,
            "spend_sats": spend_sats,
            "change_addr": outputs_described["change_addr"],
            "change_sats": outputs_described["change_sats"],
            "spend_addr": spend_addr,
            "is_batch_tx": is_batch_tx,
            # Input/output level
            "inputs_desc": inputs_described["inputs_desc"],
            "outputs_desc": outputs_described["outputs_desc"],
            "root_paths": inputs_described["root_paths_for_signing"],
        }

    def describe_p2pkh_sweep(self, privkey_obj=None):
        """
        Describe a single key p2pkh sweep transaction in a human-readable way for manual verification before signing.
        This would typically be used to sweep an old paper wallet.

        This tool supports transactions with the following constraints:
        * ALL inputs must be p2pkh and have the same private key. This does not currently support p2wpkh (TODO!).
        * There can only be 1 output (sweep transaction). Single sig is dangerous, this should only be used to migrate away.

        Note that for p2pkh we cannot directly verify the transaction fee, as the input amounts are not explicitly part of what we are prompted to sign.
        This "bug" was fixed with segwit.
        """
        self.validate()

        tx_fee_sats = self.tx_obj.fee()

        if privkey_obj:
            h160s = {
                privkey_obj.point.hash160(compressed=True),
                privkey_obj.point.hash160(compressed=False),
            }
        else:
            h160s = {}

        # Gather TX info and validate
        inputs_desc = []
        for cnt, psbt_in in enumerate(self.psbt_ins):
            psbt_in.validate()

            if psbt_in.witness_script or psbt_in.witness:
                raise SuspiciousTransaction(f"Input #{cnt} has witness")

            if psbt_in.redeem_script:
                raise SuspiciousTransaction(f"Input #{cnt} has redeem_script")

            if psbt_in.tx_in.script_pubkey().is_p2pkh() is not True:
                raise SuspiciousTransaction(f"Input #{cnt} is not p2pkh")

            # Optional check that this is the pubkey whose privkey we intend to sign with
            if h160s and psbt_in.tx_in.script_pubkey().hash160() not in h160s:
                raise SuspiciousTransaction(
                    f"Input #{cnt} is encumbered by another public key. {psbt_in.tx_in.script_pubkey().hash160()} not in {h160s}"
                )

            psbt_prev_tx_out = psbt_in.prev_tx.tx_outs[psbt_in.tx_in.prev_index]
            input_desc = {
                "prev_txhash": psbt_in.tx_in.prev_tx.hex(),
                "prev_idx": psbt_in.tx_in.prev_index,
                "n_sequence": psbt_in.tx_in.sequence,
                "sats": psbt_prev_tx_out.amount,
                "addr": psbt_prev_tx_out.script_pubkey.address(network=self.network),
            }
            inputs_desc.append(input_desc)

        total_input_sats = sum([x["sats"] for x in inputs_desc])

        if len(self.tx_obj.tx_outs) != len(self.psbt_outs):
            raise SuspiciousTransaction("Invalid PSBT: output mismatch")

        if len(self.psbt_outs) != 1:
            raise SuspiciousTransaction(
                f"PSBT with {len(self.psbt_outs)} outputs is incompatible with this script: this script only supports sweep transactions (1 output)"
            )

        psbt_out = self.psbt_outs[0]  # this script only accepts 1 output
        psbt_out.validate()

        spend_addr = psbt_out.tx_out.script_pubkey.address(network=self.network)
        output_spend_sats = psbt_out.tx_out.amount
        output_addr_type = psbt_out.tx_out.script_pubkey.__class__.__name__.rstrip(
            "ScriptPubKey"
        )
        outputs_desc = [
            {
                "sats": output_spend_sats,
                "addr": spend_addr,
                "addr_type": output_addr_type,
            }
        ]

        # comma separating satoshis for better display
        tx_summary_text = f"PSBT sends {output_spend_sats:,} sats to {spend_addr} with an UNVERIFIED fee of {tx_fee_sats:,} sats ({round(tx_fee_sats / total_input_sats * 100, 2)}% of spend)"

        return {
            # TX level:
            "tx_summary_text": tx_summary_text,
            "tx_size_bytes": len(
                self.tx_obj.serialize()
            ),  # no need to worry about vBytes because this is p2pkh
            "is_rbf_able": self.tx_obj.is_rbf_able(),
            "locktime": self.tx_obj.locktime,
            "version": self.tx_obj.version,
            "network": self.network,
            "tx_fee_sats": tx_fee_sats,
            "total_input_sats": total_input_sats,
            "output_spend_sats": output_spend_sats,
            "spend_addr": spend_addr,
            # Input/output level
            "inputs_desc": inputs_desc,
            "outputs_desc": outputs_desc,
        }


class PSBTIn:
    def __init__(
        self,
        tx_in,
        prev_tx=None,
        prev_out=None,
        sigs=None,
        hash_type=None,
        redeem_script=None,
        witness_script=None,
        named_pubs=None,
        script_sig=None,
        witness=None,
        extra_map=None,
    ):
        self.tx_in = tx_in
        self.prev_tx = prev_tx
        self.prev_out = prev_out
        self.sigs = sigs or {}
        self.hash_type = hash_type
        self.redeem_script = redeem_script
        self.witness_script = witness_script
        self.named_pubs = named_pubs or {}
        self.script_sig = script_sig
        self.witness = witness
        self.extra_map = extra_map or {}
        self.validate()

    def validate(self):
        """Checks the PSBTIn for consistency"""
        script_pubkey = self.script_pubkey()
        if self.prev_tx:
            if self.tx_in.prev_tx != self.prev_tx.hash():
                raise ValueError(
                    "previous transaction specified, but does not match the input tx"
                )
            if self.tx_in.prev_index >= len(self.prev_tx.tx_outs):
                raise ValueError("input refers to an output index that does not exist")
        if self.prev_out:
            # witness input
            if not (
                script_pubkey.is_p2sh()
                or script_pubkey.is_p2wsh()
                or script_pubkey.is_p2wpkh()
            ):
                raise ValueError("Witness UTXO provided for non-witness input")
            if self.witness_script:  # p2wsh or p2sh-p2wsh
                if not script_pubkey.is_p2wsh() and not (
                    self.redeem_script and self.redeem_script.is_p2wsh()
                ):
                    raise ValueError(
                        "WitnessScript provided for non-p2wsh ScriptPubKey"
                    )
                if self.redeem_script:
                    h160 = script_pubkey.commands[1]
                    if self.redeem_script.hash160() != h160:
                        raise ValueError(
                            "RedeemScript hash160 and ScriptPubKey hash160 do not match"
                        )
                    s256 = self.redeem_script.commands[1]
                else:
                    s256 = self.prev_out.script_pubkey.commands[1]
                if self.witness_script.sha256() != s256:
                    raise ValueError(
                        "WitnessScript sha256 and output sha256 do not match"
                    )
                for sec in self.named_pubs.keys():
                    try:
                        # this will raise a ValueError if it's not in there
                        self.witness_script.commands.index(sec)
                    except ValueError:
                        raise ValueError(
                            "pubkey is not in WitnessScript: {}".format(self)
                        )
            elif script_pubkey.is_p2wpkh() or (
                self.redeem_script and self.redeem_script.is_p2wpkh()
            ):
                if len(self.named_pubs) > 1:
                    raise ValueError("too many pubkeys in p2wpkh or p2sh-p2wpkh")
                elif len(self.named_pubs) == 1:
                    named_pub = list(self.named_pubs.values())[0]
                    if script_pubkey.commands[1] != named_pub.hash160():
                        raise ValueError(
                            "pubkey {} does not match the hash160".format(
                                named_pub.sec().hex()
                            )
                        )
        else:
            # non-witness input
            if self.redeem_script:
                if not script_pubkey.is_p2sh():
                    raise ValueError("RedeemScript defined for non-p2sh ScriptPubKey")
                # non-witness p2sh
                if self.redeem_script.is_p2wsh() or self.redeem_script.is_p2wpkh():
                    raise ValueError("Non-witness UTXO provided for witness input")
                h160 = script_pubkey.commands[1]
                if self.redeem_script.hash160() != h160:
                    raise ValueError(
                        "RedeemScript hash160 and ScriptPubKey hash160 do not match"
                    )
                for sec in self.named_pubs.keys():
                    try:
                        # this will raise a ValueError if it's not in there
                        self.redeem_script.commands.index(sec)
                    except ValueError:
                        raise ValueError(
                            "pubkey is not in RedeemScript {}".format(self)
                        )
            elif script_pubkey and script_pubkey.is_p2pkh():
                if len(self.named_pubs) > 1:
                    raise ValueError("too many pubkeys in p2pkh")
                elif len(self.named_pubs) == 1:
                    named_pub = list(self.named_pubs.values())[0]
                    if script_pubkey.commands[2] != named_pub.hash160():
                        raise ValueError(
                            "pubkey {} does not match the hash160".format(
                                named_pub.sec().hex()
                            )
                        )

    def __repr__(self):
        return "TxIn:\n{}\nPrev Tx:\n{}\nPrev Output\n{}\nSigs:\n{}\nRedeemScript:\n{}\nWitnessScript:\n{}\nPSBT Pubs:\n{}\nScriptSig:\n{}\nWitness:\n{}\n".format(
            self.tx_in,
            self.prev_tx,
            self.prev_out,
            self.sigs,
            self.redeem_script,
            self.witness_script,
            self.named_pubs,
            self.script_sig,
            self.witness,
        )

    @classmethod
    def parse(cls, s, tx_in, network=None):
        prev_tx = None
        prev_out = None
        sigs = {}
        hash_type = None
        redeem_script = None
        witness_script = None
        named_pubs = {}
        script_sig = None
        witness = None
        extra_map = {}
        key = read_varstr(s)
        while key != b"":
            psbt_type = key[0:1]
            if psbt_type == PSBT_IN_NON_WITNESS_UTXO:
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if prev_tx:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                tx_len = read_varint(s)
                prev_tx = Tx.parse(s)
                if len(prev_tx.serialize()) != tx_len:
                    raise IOError("tx length does not match")
                tx_in._value = prev_tx.tx_outs[tx_in.prev_index].amount
                tx_in._script_pubkey = prev_tx.tx_outs[tx_in.prev_index].script_pubkey
            elif psbt_type == PSBT_IN_WITNESS_UTXO:
                tx_out_len = read_varint(s)
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if prev_out:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                prev_out = TxOut.parse(s)
                if len(prev_out.serialize()) != tx_out_len:
                    raise ValueError("tx out length does not match")
                tx_in._value = prev_out.amount
                tx_in._script_pubkey = prev_out.script_pubkey
            elif psbt_type == PSBT_IN_PARTIAL_SIG:
                if sigs.get(key[1:]):
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                sigs[key[1:]] = read_varstr(s)
            elif psbt_type == PSBT_IN_SIGHASH_TYPE:
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if hash_type:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                hash_type = little_endian_to_int(read_varstr(s))
            elif psbt_type == PSBT_IN_REDEEM_SCRIPT:
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if redeem_script:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                redeem_script = RedeemScript.parse(s)
            elif psbt_type == PSBT_IN_WITNESS_SCRIPT:
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if witness_script:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                witness_script = WitnessScript.parse(s)
            elif psbt_type == PSBT_IN_BIP32_DERIVATION:
                if len(key) != 34:
                    raise KeyError("Wrong length for the key")
                named_pub = NamedPublicKey.parse(key, s, network=network)
                named_pubs[named_pub.sec()] = named_pub
            elif psbt_type == PSBT_IN_FINAL_SCRIPTSIG:
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if script_sig:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                script_sig = Script.parse(s)
            elif psbt_type == PSBT_IN_FINAL_SCRIPTWITNESS:
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if witness:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                _ = read_varint(s)
                witness = Witness.parse(s)
            else:
                if extra_map.get(key):
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                extra_map[key] = read_varstr(s)
            key = read_varstr(s)
        return cls(
            tx_in,
            prev_tx,
            prev_out,
            sigs,
            hash_type,
            redeem_script,
            witness_script,
            named_pubs,
            script_sig,
            witness,
            extra_map,
        )

    def serialize(self):
        result = b""
        if self.prev_tx:
            result += serialize_key_value(
                PSBT_IN_NON_WITNESS_UTXO, self.prev_tx.serialize()
            )
        elif self.prev_out:
            result += serialize_key_value(
                PSBT_IN_WITNESS_UTXO, self.prev_out.serialize()
            )
        # we need to put the keys in the witness script or redeem script order
        keys = []
        if self.witness_script:
            for command in self.witness_script.commands:
                if self.sigs.get(command):
                    keys.append(command)
        elif self.redeem_script and not self.redeem_script.is_p2wpkh():
            for command in self.redeem_script.commands:
                if self.sigs.get(command):
                    keys.append(command)
        else:
            keys = sorted(self.sigs.keys())
        for key in keys:
            result += serialize_key_value(PSBT_IN_PARTIAL_SIG + key, self.sigs[key])
        if self.hash_type:
            result += serialize_key_value(
                PSBT_IN_SIGHASH_TYPE, int_to_little_endian(self.hash_type, 4)
            )
        if self.redeem_script:
            result += serialize_key_value(
                PSBT_IN_REDEEM_SCRIPT, self.redeem_script.raw_serialize()
            )
        if self.witness_script:
            result += serialize_key_value(
                PSBT_IN_WITNESS_SCRIPT, self.witness_script.raw_serialize()
            )
        for sec in sorted(self.named_pubs.keys()):
            named_pub = self.named_pubs[sec]
            result += named_pub.serialize(PSBT_IN_BIP32_DERIVATION)
        if self.script_sig:
            result += serialize_key_value(
                PSBT_IN_FINAL_SCRIPTSIG, self.script_sig.raw_serialize()
            )
        if self.witness:
            result += serialize_key_value(
                PSBT_IN_FINAL_SCRIPTWITNESS, self.witness.serialize()
            )
        # extra
        for key in sorted(self.extra_map.keys()):
            result += encode_varstr(key) + encode_varstr(self.extra_map[key])
        # delimiter
        result += PSBT_DELIMITER
        return result

    def script_pubkey(self):
        if self.prev_tx:
            return self.prev_tx.tx_outs[self.tx_in.prev_index].script_pubkey
        elif self.prev_out:
            return self.prev_out.script_pubkey
        else:
            return None

    def use_segwit_signature(self):
        # https://github.com/buidl-bitcoin/buidl-python/issues/23
        if self.witness_script or self.witness:
            return True
        if self.redeem_script and self.redeem_script.is_witness_script():
            return True
        if self.script_pubkey() and self.script_pubkey().is_witness_script():
            return True
        return False

    def update(self, tx_lookup, pubkey_lookup, redeem_lookup, witness_lookup):
        """Updates the input with NamedPublicKeys, RedeemScript or WitnessScript that
        correspond"""
        # the input might already have a previous transaction
        prev_tx = self.prev_tx or tx_lookup.get(self.tx_in.prev_tx)
        # grab the output at the previous index, or alternatively get the self.prev_out
        if prev_tx:
            prev_out = prev_tx.tx_outs[self.tx_in.prev_index]
        else:
            prev_out = self.prev_out
        # if we don't know the previous output we can't update anything
        if not prev_tx and not prev_out:
            return
        # get the ScriptPubKey that we're unlocking
        script_pubkey = prev_out.script_pubkey
        # Set the _value and _script_pubkey properties of the TxIn object
        #  so that no full node is needed to look those up
        self.tx_in._value = prev_out.amount
        self.tx_in._script_pubkey = script_pubkey
        # grab the RedeemScript
        if script_pubkey.is_p2sh():
            # see if we have a RedeemScript already defined or in the lookup
            self.redeem_script = self.redeem_script or redeem_lookup.get(
                script_pubkey.commands[1]
            )
            # if there's no RedeemScript, we can't do any more updating, so return
            if not self.redeem_script:
                return
        # Exercise 2: if we have p2wpkh or p2sh-p2wpkh see if we have the appropriate NamedPublicKey
        if script_pubkey.is_p2wpkh() or (
            self.redeem_script and self.redeem_script.is_p2wpkh()
        ):
            # set the prev_out property as this is Segwit
            self.prev_out = prev_out
            # for p2wpkh, the hash160 is the second command of the ScriptPubKey
            # for p2sh-p2wpkh, the hash160 is the second command of the RedeemScript
            if script_pubkey.is_p2wpkh():
                h160 = script_pubkey.commands[1]
            else:
                h160 = self.redeem_script.commands[1]
            # see if we have the public key that corresponds to the hash160
            named_pub = pubkey_lookup.get(h160)
            # if so add it to the named_pubs dictionary
            if named_pub:
                self.named_pubs[named_pub.sec()] = named_pub.point
        # Exercise 12: if we have p2wsh or p2sh-p2wsh see if we have one or more NamedPublicKeys
        elif script_pubkey.is_p2wsh() or (
            self.redeem_script and self.redeem_script.is_p2wsh()
        ):
            # set the prev_out property as this is Segwit
            self.prev_out = prev_out
            # for p2wsh, the sha256 is the second command of the ScriptPubKey
            # for p2sh-p2wsh, the sha256 is the second command of the RedeemScript
            if script_pubkey.is_p2wsh():
                s256 = script_pubkey.commands[1]
            else:
                s256 = self.redeem_script.commands[1]
            # see if we have the WitnessScript that corresponds to the sha256
            self.witness_script = self.witness_script or witness_lookup.get(s256)
            if self.witness_script:
                # go through the commands of the WitnessScript for NamedPublicKeys
                for command in self.witness_script.commands:
                    named_pub = pubkey_lookup.get(command)
                    if named_pub:
                        self.named_pubs[named_pub.sec()] = named_pub.point
        # we've eliminated p2sh wrapped segwit, handle p2sh here
        elif script_pubkey.is_p2sh():
            # set the prev_tx property as it's not segwit
            self.prev_tx = prev_tx
            # go through the commands of the RedeemScript for NamedPublicKeys
            for command in self.redeem_script.commands:
                # if we find a NamedPublicKey, add to the named_pubs dictionary
                #  key is compressed sec, value is the point object
                named_pub = pubkey_lookup.get(command)
                if named_pub:
                    self.named_pubs[named_pub.sec()] = named_pub.point
        # if we have p2pkh, see if we have the appropriate NamedPublicKey
        elif script_pubkey.is_p2pkh():
            # set the prev_tx property as it's not segwit
            self.prev_tx = prev_tx
            # look for the NamedPublicKey that corresponds to the hash160
            #  which is the 3rd command of the ScriptPubKey
            named_pub = pubkey_lookup.get(script_pubkey.commands[2])
            if named_pub:
                # if it exists, add to the named_pubs dict
                #  key is the sec and the value is the point
                self.named_pubs[named_pub.sec()] = named_pub.point
        # else we throw a ValueError
        else:
            raise ValueError(
                "cannot update a transaction because it is not p2pkh, p2sh, p2wpkh or p2wsh: {}".format(
                    script_pubkey
                )
            )

    def combine(self, other):
        """Combines two PSBTIn objects into self"""
        # if prev_tx is defined in the other, but not in self, add
        if self.prev_tx is None and other.prev_tx:
            self.prev_tx = other.prev_tx
        # if prev_tx is defined in the other, but not in self, add
        if self.prev_out is None and other.prev_out:
            self.prev_out = other.prev_out
        # combine the sigs
        self.sigs = {**self.sigs, **other.sigs}
        # if hash_type is defined in the other, but not in self, add
        if self.hash_type is None and other.hash_type:
            self.hash_type = other.hash_type
        # if redeem_script is defined in the other, but not in self, add
        if self.redeem_script is None and other.redeem_script:
            self.redeem_script = other.redeem_script
        # if witness_script is defined in the other, but not in self, add
        if self.witness_script is None and other.witness_script:
            self.witness_script = other.witness_script
        # combine the pubs
        self.named_pubs = {**other.named_pubs, **self.named_pubs}
        # if script_sig is defined in the other, but not in self, add
        if self.script_sig is None and other.script_sig:
            self.script_sig = other.script_sig
        # if witness is defined in the other, but not in self, add
        if self.witness is None and other.witness:
            self.witness = other.witness
        # combine extra_map
        self.extra_map = {**other.extra_map, **self.extra_map}

    def finalize(self):
        """Removes all sigs/named pubs/RedeemScripts/WitnessScripts and
        sets the script_sig and witness fields"""
        # get the ScriptPubKey for this input
        script_pubkey = self.script_pubkey()
        # if the ScriptPubKey is p2sh
        if script_pubkey.is_p2sh():
            # make sure there's a RedeemScript
            if not self.redeem_script:
                raise RuntimeError("Cannot finalize p2sh without a RedeemScript")
        # Exercise 6: if p2wpkh or p2sh-p2wpkh
        if script_pubkey.is_p2wpkh() or (
            self.redeem_script and self.redeem_script.is_p2wpkh()
        ):
            # check to see that we have exactly 1 signature
            if len(self.sigs) != 1:
                raise RuntimeError(
                    "p2wpkh or p2sh-p2wpkh should have exactly 1 signature"
                )
            # the key of the sigs dict is the compressed SEC pubkey
            sec = list(self.sigs.keys())[0]
            # the value of the sigs dict is the signature
            sig = list(self.sigs.values())[0]
            # set the ScriptSig to the RedeemScript if there is one
            if self.redeem_script:
                self.script_sig = Script([self.redeem_script.raw_serialize()])
            else:
                self.script_sig = Script()
            # set the Witness to sig and sec
            self.witness = Witness([sig, sec])
        # Exercise 15: if p2wsh or p2sh-p2wsh
        elif script_pubkey.is_p2wsh() or (
            self.redeem_script and self.redeem_script.is_p2wsh()
        ):
            # make sure there's a WitnessScript
            if not self.witness_script:
                raise RuntimeError(
                    "Cannot finalize p2wsh or p2sh-p2wsh without a WitnessScript"
                )
            # convert the first command to a number (required # of sigs)
            num_sigs = op_code_to_number(self.witness_script.commands[0])
            # make sure we have at least the number of sigs required
            if len(self.sigs) < num_sigs:
                raise RuntimeError(
                    "Cannot finalize p2wsh or p2sh-p2wsh because {} sigs were provided where {} were needed".format(
                        len(self.sigs), num_sigs
                    )
                )
            # create a list of items for the Witness. Start with b'\x00' for the
            #  OP_CHECKMULTISIG off-by-one error
            witness_items = [b"\x00"]
            # for each command in the WitnessScript
            for command in self.witness_script.commands:
                # grab the sig for the pubkey
                sig = self.sigs.get(command)
                # if the sig exists, then add to the Witness item list
                if sig is not None:
                    witness_items.append(sig)
                # when we have enough signatures, break
                if len(witness_items) - 1 >= num_sigs:
                    break
            # make sure we have enough sigs to pass validation
            if len(witness_items) - 1 < num_sigs:
                raise RuntimeError("Not enough signatures provided for p2sh-p2wsh")
            # add the raw WitnessScript as the last item for p2wsh execution
            witness_items.append(self.witness_script.raw_serialize())
            # create the witness
            self.witness = Witness(witness_items)
            # set the ScriptSig to the RedeemScript if there is one
            if self.redeem_script:
                self.script_sig = Script([self.redeem_script.raw_serialize()])
            else:
                self.script_sig = Script()
        # we've eliminated p2sh wrapped segwit, handle p2sh here
        elif script_pubkey.is_p2sh():
            # convert the first command to a number (required # of sigs)
            num_sigs = op_code_to_number(self.redeem_script.commands[0])
            # make sure we have at least the number of sigs required
            if len(self.sigs) < num_sigs:
                raise RuntimeError(
                    "Cannot finalize p2sh because {} sigs were provided where {} were needed".format(
                        len(self.sigs), num_sigs
                    )
                )
            # create a list of commands for the ScriptSig. Start with 0 for the
            #  OP_CHECKMULTISIG off-by-one error
            script_sig_commands = [0]
            # for each command in the RedeemScript
            for command in self.redeem_script.commands:
                # skip if the command is an integer
                if type(command) == int:
                    continue
                # grab the sig for the pubkey
                sig = self.sigs.get(command)
                # if the sig exists, then add to the ScriptSig command list
                if sig is not None:
                    script_sig_commands.append(sig)
                # when we have enough signatures, break
                if len(script_sig_commands) - 1 >= num_sigs:
                    break
            # make sure we have enough sigs to pass validation
            if len(script_sig_commands) < num_sigs:
                raise RuntimeError("Not enough signatures provided for p2wsh")
            # add the raw redeem script as the last command for p2sh execution
            script_sig_commands.append(self.redeem_script.raw_serialize())
            # change the ScriptSig to be a Script with the commands we've gathered
            self.script_sig = Script(script_sig_commands)
        elif script_pubkey.is_p2pkh():
            # check to see that we have exactly 1 signature
            if len(self.sigs) != 1:
                raise RuntimeError("P2pkh requires exactly 1 signature")
            # the key of the sigs dict is the compressed SEC pubkey
            sec = list(self.sigs.keys())[0]
            # the value of the sigs dict is the signature
            sig = list(self.sigs.values())[0]
            # set the ScriptSig, which is Script([sig, sec])
            self.script_sig = Script([sig, sec])
        else:
            raise ValueError(
                "Cannot finalize this ScriptPubKey: {}".format(script_pubkey)
            )
        # reset sigs, hash_type, redeem_script, witness_script and named_pubs to be empty
        self.sigs = {}
        self.hash_type = None
        self.redeem_script = None
        self.witness_script = None
        self.named_pubs = {}


class PSBTOut:
    def __init__(
        self,
        tx_out,
        redeem_script=None,
        witness_script=None,
        named_pubs=None,
        extra_map=None,
    ):
        self.tx_out = tx_out
        self.redeem_script = redeem_script
        self.witness_script = witness_script
        self.named_pubs = named_pubs or {}
        self.extra_map = extra_map or {}
        self.validate()

    def validate(self):
        """Checks the PSBTOut for consistency"""
        script_pubkey = self.tx_out.script_pubkey
        if script_pubkey.is_p2pkh():
            if self.redeem_script:
                raise KeyError("RedeemScript included in p2pkh output")
            if self.witness_script:
                raise KeyError("WitnessScript included in p2pkh output")
            if len(self.named_pubs) > 1:
                raise ValueError("too many pubkeys in p2pkh")
            elif len(self.named_pubs) == 1:
                named_pub = list(self.named_pubs.values())[0]
                if script_pubkey.commands[2] != named_pub.hash160():
                    raise ValueError(
                        "pubkey {} does not match the hash160".format(
                            named_pub.sec().hex()
                        )
                    )
        elif script_pubkey.is_p2wpkh():
            if self.redeem_script:
                raise KeyError("RedeemScript included in p2wpkh output")
            if self.witness_script:
                raise KeyError("WitnessScript included in p2wpkh output")
            if len(self.named_pubs) > 1:
                raise ValueError("too many pubkeys in p2wpkh")
            elif len(self.named_pubs) == 1:
                named_pub = list(self.named_pubs.values())[0]
                if script_pubkey.commands[1] != named_pub.hash160():
                    raise ValueError(
                        "pubkey {} does not match the hash160".format(
                            named_pub.sec().hex()
                        )
                    )
        elif self.witness_script:
            if self.redeem_script:
                h160 = script_pubkey.commands[1]
                if self.redeem_script.hash160() != h160:
                    raise ValueError(
                        "RedeemScript hash160 and ScriptPubKey hash160 do not match"
                    )
                s256 = self.redeem_script.commands[1]
            else:
                s256 = script_pubkey.commands[1]
            if self.witness_script.sha256() != s256:
                raise ValueError(
                    "WitnessScript sha256 and output sha256 do not match {} {}".format(
                        self, self.witness_script.sha256().hex()
                    )
                )
            for sec in self.named_pubs.keys():
                try:
                    # this will raise a ValueError if it's not in there
                    self.witness_script.commands.index(sec)
                except ValueError:
                    raise ValueError("pubkey is not in WitnessScript {}".format(self))
        elif self.redeem_script:
            for sec in self.named_pubs.keys():
                try:
                    # this will raise a ValueError if it's not in there
                    self.redeem_script.commands.index(sec)
                except ValueError:
                    raise ValueError("pubkey is not in RedeemScript {}".format(self))

    def __repr__(self):
        return (
            "TxOut:\n{}\nRedeemScript:\n{}\nWitnessScript\n{}\nPSBT Pubs:\n{}\n".format(
                self.tx_out, self.redeem_script, self.witness_script, self.named_pubs
            )
        )

    @classmethod
    def parse(cls, s, tx_out, network=None):
        redeem_script = None
        witness_script = None
        named_pubs = {}
        extra_map = {}
        key = read_varstr(s)
        while key != b"":
            psbt_type = key[0:1]
            if psbt_type == PSBT_OUT_REDEEM_SCRIPT:
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if redeem_script:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                redeem_script = RedeemScript.parse(s)
            elif psbt_type == PSBT_OUT_WITNESS_SCRIPT:
                if len(key) != 1:
                    raise KeyError("Wrong length for the key")
                if witness_script:
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                witness_script = WitnessScript.parse(s)
            elif psbt_type == PSBT_OUT_BIP32_DERIVATION:
                if len(key) != 34:
                    raise KeyError("Wrong length for the key")
                named_pub = NamedPublicKey.parse(key, s, network=network)
                named_pubs[named_pub.sec()] = named_pub
            else:
                if extra_map.get(key):
                    raise KeyError("Duplicate Key in parsing: {}".format(key.hex()))
                extra_map[key] = read_varstr(s)
            key = read_varstr(s)
        return cls(tx_out, redeem_script, witness_script, named_pubs, extra_map)

    def serialize(self):
        result = b""
        if self.redeem_script:
            result += serialize_key_value(
                PSBT_OUT_REDEEM_SCRIPT, self.redeem_script.raw_serialize()
            )
        if self.witness_script:
            result += serialize_key_value(
                PSBT_OUT_WITNESS_SCRIPT, self.witness_script.raw_serialize()
            )
        for key in sorted(self.named_pubs.keys()):
            named_pub = self.named_pubs[key]
            result += named_pub.serialize(PSBT_OUT_BIP32_DERIVATION)
        # extra
        for key in sorted(self.extra_map.keys()):
            result += encode_varstr(key) + encode_varstr(self.extra_map[key])
        # delimiter
        result += PSBT_DELIMITER
        return result

    def update(self, pubkey_lookup, redeem_lookup, witness_lookup):
        """Updates the output with NamedPublicKeys, RedeemScript or WitnessScript that
        correspond"""
        # get the ScriptPubKey
        script_pubkey = self.tx_out.script_pubkey
        # if the ScriptPubKey is p2sh, check for a RedeemScript
        if script_pubkey.is_p2sh():
            self.redeem_script = redeem_lookup.get(script_pubkey.commands[1])
            # if no RedeemScript exists, we can't update, so return
            if not self.redeem_script:
                return
        # Exercise 2: if p2wpkh or p2sh-p2wpkh
        if script_pubkey.is_p2wpkh() or (
            self.redeem_script and self.redeem_script.is_p2wpkh()
        ):
            # get the hash160 (second command of RedeemScript or ScriptPubKey)
            if self.redeem_script:
                h160 = self.redeem_script.commands[1]
            else:
                h160 = script_pubkey.commands[1]
            # look for the NamedPublicKey and add if there
            named_pub = pubkey_lookup.get(h160)
            if named_pub:
                self.named_pubs[named_pub.sec()] = named_pub.point
        # Exercise 12: if p2wsh/p2sh-p2wsh
        elif script_pubkey.is_p2wsh() or (
            self.redeem_script and self.redeem_script.is_p2wsh()
        ):
            # get the sha256 (second command of RedeemScript or ScriptPubKey)
            if self.redeem_script:
                s256 = self.redeem_script.commands[1]
            else:
                s256 = script_pubkey.commands[1]
            # look for the WitnessScript using the sha256
            witness_script = witness_lookup.get(s256)
            if witness_script:
                # update the WitnessScript
                self.witness_script = witness_script
                # look through the WitnessScript for any NamedPublicKeys
                for command in witness_script.commands:
                    named_pub = pubkey_lookup.get(command)
                    # if found, add the NamedPublicKey
                    if named_pub:
                        self.named_pubs[named_pub.sec()] = named_pub.point
        # we've eliminated p2sh wrapped segwit, handle p2sh here
        elif script_pubkey.is_p2sh():
            # Look through the commands in the RedeemScript for any NamedPublicKeys
            for command in self.redeem_script.commands:
                named_pub = pubkey_lookup.get(command)
                # if a NamedPublicKey exists
                if named_pub:
                    # add to the named_pubs dictionary
                    #  key is sec and the point is the value
                    self.named_pubs[named_pub.sec()] = named_pub.point
        # Exercise 3: if the ScriptPubKey is p2pkh,
        elif script_pubkey.is_p2pkh():
            # Look at the third command of the ScriptPubKey for the hash160
            # Use that to look up the NamedPublicKey
            named_pub = pubkey_lookup.get(script_pubkey.commands[2])
            # if a NamedPublicKey exists
            if named_pub:
                # add to the named_pubs dictionary
                #  key is sec and the point is the value
                self.named_pubs[named_pub.sec()] = named_pub.point

    def combine(self, other):
        """Combines two PSBTOuts to self"""
        # if redeem_script is defined in the other, but not in self, add
        if self.redeem_script is None and other.redeem_script:
            self.redeem_script = other.redeem_script
        # if witness_script is defined in the other, but not in self, add
        if self.witness_script is None and other.witness_script:
            self.witness_script = other.witness_script
        # combine the pubs
        self.named_pubs = {**other.named_pubs, **self.named_pubs}
        # combine extra_map
        self.extra_map = {**other.extra_map, **self.extra_map}



import socket
import time

from io import BytesIO
from random import randint
from time import sleep

from buidl.block import Block
from buidl.helper import (
    GOLOMB_M,
    GOLOMB_P,
    decode_golomb,
    encode_varint,
    hash256,
    hash_to_range,
    int_to_little_endian,
    little_endian_to_int,
    read_varint,
    unpack_bits,
)
from buidl.tx import Tx

TX_DATA_TYPE = 1
BLOCK_DATA_TYPE = 2
FILTERED_BLOCK_DATA_TYPE = 3
COMPACT_BLOCK_DATA_TYPE = 4
WITNESS_TX_DATA_TYPE = (1 << 30) + TX_DATA_TYPE
WITNESS_BLOCK_DATA_TYPE = (1 << 30) + BLOCK_DATA_TYPE

MAGIC = {
    "mainnet": b"\xf9\xbe\xb4\xd9",
    "testnet": b"\x0b\x11\x09\x07",
    "signet": b"\x0a\x03\xcf\x40",
}
PORT = {
    "mainnet": 8333,
    "testnet": 18333,
    "signet": 38333,
}


BASIC_FILTER_TYPE = 0


class NetworkEnvelope:
    def __init__(self, command, payload, network="mainnet"):
        self.command = command
        self.payload = payload
        self.magic = MAGIC[network]

    def __repr__(self):
        return "{}: {}".format(
            self.command.decode("ascii"),
            self.payload.hex(),
        )

    @classmethod
    def parse(cls, s, network="mainnet"):
        """Takes a stream and creates a NetworkEnvelope"""
        # check the network magic
        magic = s.read(4)
        if magic == b"":
            raise RuntimeError("Connection reset!")
        expected_magic = MAGIC[network]
        if magic != expected_magic:
            raise RuntimeError(
                "magic is not right {} vs {}".format(magic.hex(), expected_magic.hex())
            )
        # command 12 bytes, strip the trailing 0's using .strip(b'\x00')
        command = s.read(12).strip(b"\x00")
        # payload length 4 bytes, little endian
        payload_length = little_endian_to_int(s.read(4))
        # checksum 4 bytes, first four of hash256 of payload
        checksum = s.read(4)
        # payload is of length payload_length
        payload = s.read(payload_length)
        # verify checksum
        calculated_checksum = hash256(payload)[:4]
        if calculated_checksum != checksum:
            raise RuntimeError("checksum does not match")
        return cls(command, payload, network=network)

    def serialize(self):
        """Returns the byte serialization of the entire network message"""
        # add the network magic using self.magic
        result = self.magic
        # command 12 bytes, fill leftover with b'\x00' * (12 - len(self.command))
        result += self.command + b"\x00" * (12 - len(self.command))
        # payload length 4 bytes, little endian
        result += int_to_little_endian(len(self.payload), 4)
        # checksum 4 bytes, first four of hash256 of payload
        result += hash256(self.payload)[:4]
        # payload
        result += self.payload
        return result

    def stream(self):
        """Returns a stream for parsing the payload"""
        return BytesIO(self.payload)


class VersionMessage:
    command = b"version"

    def __init__(
        self,
        version=70015,
        services=0,
        timestamp=None,
        receiver_services=0,
        receiver_ip=b"\x00\x00\x00\x00",
        receiver_port=8333,
        sender_services=0,
        sender_ip=b"\x00\x00\x00\x00",
        sender_port=8333,
        nonce=None,
        user_agent=b"/programmingblockchain:0.1/",
        latest_block=0,
        relay=True,
    ):
        self.version = version
        self.services = services
        if timestamp is None:
            self.timestamp = int(time.time())
        else:
            self.timestamp = timestamp
        self.receiver_services = receiver_services
        self.receiver_ip = receiver_ip
        self.receiver_port = receiver_port
        self.sender_services = sender_services
        self.sender_ip = sender_ip
        self.sender_port = sender_port
        if nonce is None:
            self.nonce = int_to_little_endian(randint(0, 2 ** 64), 8)
        else:
            self.nonce = nonce
        self.user_agent = user_agent
        self.latest_block = latest_block
        self.relay = relay

    def serialize(self):
        """Serialize this message to send over the network"""
        # version is 4 bytes little endian
        result = int_to_little_endian(self.version, 4)
        # services is 8 bytes little endian
        result += int_to_little_endian(self.services, 8)
        # timestamp is 8 bytes little endian
        result += int_to_little_endian(self.timestamp, 8)
        # receiver services is 8 bytes little endian
        result += int_to_little_endian(self.receiver_services, 8)
        # IPV4 is 10 00 bytes and 2 ff bytes then receiver ip
        result += b"\x00" * 10 + b"\xff\xff" + self.receiver_ip
        # receiver port is 2 bytes, little endian
        result += int_to_little_endian(self.receiver_port, 2)
        # sender services is 8 bytes little endian
        result += int_to_little_endian(self.sender_services, 8)
        # IPV4 is 10 00 bytes and 2 ff bytes then sender ip
        result += b"\x00" * 10 + b"\xff\xff" + self.sender_ip
        # sender port is 2 bytes, little endian
        result += int_to_little_endian(self.sender_port, 2)
        # nonce
        result += self.nonce
        # useragent is a variable string, so varint first
        result += encode_varint(len(self.user_agent))
        result += self.user_agent
        # latest block is 4 bytes little endian
        result += int_to_little_endian(self.latest_block, 4)
        # relay is 00 if false, 01 if true
        if self.relay:
            result += b"\x01"
        else:
            result += b"\x00"
        return result


class VerAckMessage:
    command = b"verack"

    def __init__(self):
        pass

    @classmethod
    def parse(cls, s):
        return cls()

    def serialize(self):
        return b""


class PingMessage:
    command = b"ping"

    def __init__(self, nonce):
        self.nonce = nonce

    @classmethod
    def parse(cls, s):
        nonce = s.read(8)
        return cls(nonce)

    def serialize(self):
        return self.nonce


class PongMessage:
    command = b"pong"

    def __init__(self, nonce):
        self.nonce = nonce

    def parse(cls, s):
        nonce = s.read(8)
        return cls(nonce)

    def serialize(self):
        return self.nonce


class GetHeadersMessage:
    command = b"getheaders"

    def __init__(self, version=70015, num_hashes=1, start_block=None, end_block=None):
        self.version = version
        self.num_hashes = num_hashes
        if start_block is None:
            raise RuntimeError("a start block is required")
        self.start_block = start_block
        if end_block is None:
            self.end_block = b"\x00" * 32
        else:
            self.end_block = end_block

    def serialize(self):
        """Serialize this message to send over the network"""
        # protocol version is 4 bytes little-endian
        result = int_to_little_endian(self.version, 4)
        # number of hashes is a varint
        result += encode_varint(self.num_hashes)
        # start block is in little-endian
        result += self.start_block[::-1]
        # end block is also in little-endian
        result += self.end_block[::-1]
        return result


class HeadersMessage:
    command = b"headers"

    def __init__(self, headers):
        self.headers = headers

    @classmethod
    def parse(cls, s):
        # number of headers is in a varint
        num_headers = read_varint(s)
        # initialize the headers array
        headers = []
        # loop through number of headers times
        for _ in range(num_headers):
            # add a header to the headers array by using Block.parse_header(s)
            headers.append(Block.parse_header(s))
            # read the next varint (num_txs)
            num_txs = read_varint(s)
            # num_txs should be 0 or raise a RuntimeError
            if num_txs != 0:
                raise RuntimeError("number of txs not 0")
        # return a class instance
        return cls(headers)

    def is_valid(self):
        """Return whether the headers satisfy proof-of-work and are sequential"""
        last_block = None
        for h in self.headers:
            if not h.check_pow():
                return False
            if last_block and h.prev_block != last_block:
                return False
            last_block = h.hash()
        return True


class GetDataMessage:
    command = b"getdata"

    def __init__(self):
        self.data = []

    def add_data(self, data_type, identifier):
        self.data.append((data_type, identifier))

    def serialize(self):
        # start with the number of items as a varint
        result = encode_varint(len(self.data))
        for data_type, identifier in self.data:
            # data type is 4 bytes little endian
            result += int_to_little_endian(data_type, 4)
            # identifier needs to be in little endian
            result += identifier[::-1]
        return result


class GetCFiltersMessage:
    command = b"getcfilters"

    def __init__(self, filter_type=BASIC_FILTER_TYPE, start_height=1, stop_hash=None):
        self.filter_type = filter_type
        self.start_height = start_height
        if stop_hash is None:
            raise RuntimeError
        self.stop_hash = stop_hash

    def serialize(self):
        result = self.filter_type.to_bytes(1, "big")
        result += int_to_little_endian(self.start_height, 4)
        result += self.stop_hash[::-1]
        return result


class CFilterMessage:
    command = b"cfilter"

    def __init__(self, filter_type, block_hash, filter_bytes, hashes):
        self.filter_type = filter_type
        self.block_hash = block_hash
        self.filter_bytes = filter_bytes
        self.hashes = hashes
        self.f = len(self.hashes) * GOLOMB_M
        self.key = self.block_hash[::-1][:16]

    @classmethod
    def parse(cls, s):
        filter_type = s.read(1)[0]
        block_hash = s.read(32)[::-1]
        num_bytes = read_varint(s)
        filter_bytes = s.read(num_bytes)
        substream = BytesIO(filter_bytes)
        n = read_varint(substream)
        bits = unpack_bits(substream.read())
        hashes = set()
        current = 0
        for _ in range(n):
            delta = decode_golomb(bits, GOLOMB_P)
            current += delta
            hashes.add(current)
        return cls(filter_type, block_hash, filter_bytes, hashes)

    def hash(self, raw_script_pubkey):
        return hash_to_range(self.key, raw_script_pubkey, self.f)

    def __contains__(self, raw_script_pubkey):
        if type(raw_script_pubkey) == bytes:
            return self.hash(raw_script_pubkey) in self.hashes
        else:
            for r in raw_script_pubkey:
                if self.hash(r) in self.hashes:
                    return True
            return False


class GetCFHeadersMessage:
    command = b"getcfheaders"

    def __init__(self, filter_type=BASIC_FILTER_TYPE, start_height=0, stop_hash=None):
        self.filter_type = filter_type
        self.start_height = start_height
        if stop_hash is None:
            raise RuntimeError
        self.stop_hash = stop_hash

    def serialize(self):
        result = self.filter_type.to_bytes(1, "big")
        result += int_to_little_endian(self.start_height, 4)
        result += self.stop_hash[::-1]
        return result


class CFHeadersMessage:
    command = b"cfheaders"

    def __init__(self, filter_type, stop_hash, previous_filter_header, filter_hashes):
        self.filter_type = filter_type
        self.stop_hash = stop_hash
        self.previous_filter_header = previous_filter_header
        self.filter_hashes = filter_hashes

    @classmethod
    def parse(cls, s):
        filter_type = s.read(1)[0]
        stop_hash = s.read(32)[::-1]
        previous_filter_header = s.read(32)[::-1]
        filter_hashes_length = read_varint(s)
        filter_hashes = []
        for _ in range(filter_hashes_length):
            filter_hashes.append(s.read(32)[::-1])
        return cls(filter_type, stop_hash, previous_filter_header, filter_hashes)


class GetCFCheckPointMessage:
    command = b"getcfcheckpt"

    def __init__(self, filter_type=BASIC_FILTER_TYPE, stop_hash=None):
        self.filter_type = filter_type
        if stop_hash is None:
            raise RuntimeError("Need a stop hash")
        self.stop_hash = stop_hash

    def serialize(self):
        result = self.filter_type.to_bytes(1, "big")
        result += self.stop_hash[::-1]
        return result


class CFCheckPointMessage:
    command = b"cfcheckpt"

    def __init__(self, filter_type, stop_hash, filter_headers):
        self.filter_type = filter_type
        self.stop_hash = stop_hash
        self.filter_headers = filter_headers

    @classmethod
    def parse(cls, s):
        filter_type = s.read(1)[0]
        stop_hash = s.read(32)[::-1]
        filter_headers_length = read_varint(s)
        filter_headers = []
        for _ in range(filter_headers_length):
            filter_headers.append(s.read(32)[::-1])
        return cls(filter_type, stop_hash, filter_headers)


class GenericMessage:
    def __init__(self, command, payload):
        self.command = command
        self.payload = payload

    def serialize(self):
        return self.payload


class SimpleNode:
    def __init__(self, host, port=None, network="mainnet", logging=False):
        if port is None:
            port = PORT[network]
        self.network = network
        self.logging = logging
        # connect to socket
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((host, port))
        # create a stream that we can use with the rest of the library
        self.stream = self.socket.makefile("rb", None)

    def handshake(self):
        """Do a handshake with the other node. Handshake is sending a version message and getting a verack back."""
        # create a version message
        version = VersionMessage()
        # send the command
        self.send(version)
        # wait for a verack message
        self.wait_for(VerAckMessage)

    def send(self, message):
        """Send a message to the connected node"""
        # create a network envelope
        envelope = NetworkEnvelope(
            message.command, message.serialize(), network=self.network
        )
        if self.logging:
            print("sending: {}".format(envelope))
        # send the serialized envelope over the socket using sendall
        self.socket.sendall(envelope.serialize())

    def read(self):
        """Read a message from the socket"""
        envelope = NetworkEnvelope.parse(self.stream, network=self.network)
        if self.logging:
            print("receiving: {}".format(envelope))
        return envelope

    def wait_for(self, *message_classes):
        """Wait for one of the messages in the list"""
        # initialize the command we have, which should be None
        command = None
        command_to_class = {m.command: m for m in message_classes}
        # loop until the command is in the commands we want
        while command not in command_to_class.keys():
            # get the next network message
            envelope = self.read()
            # set the command to be evaluated
            command = envelope.command
            # we know how to respond to version and ping, handle that here
            if command == VersionMessage.command:
                # send verack
                self.send(VerAckMessage())
            elif command == PingMessage.command:
                # send pong
                self.send(PongMessage(envelope.payload))
        # return the envelope parsed as a member of the right message class
        return command_to_class[command].parse(envelope.stream())

    def get_filtered_txs(self, block_hashes):
        """Returns transactions that match the bloom filter"""
        from buidl.merkleblock import MerkleBlock

        # create a getdata message
        getdata = GetDataMessage()
        # for each block request the filtered block
        for block_hash in block_hashes:
            # add_data (FILTERED_BLOCK_DATA_TYPE, block_hash) to request the block
            getdata.add_data(FILTERED_BLOCK_DATA_TYPE, block_hash)
        # send the getdata message
        self.send(getdata)
        # initialize the results array we'll send back
        results = []
        # for each block hash
        for block_hash in block_hashes:
            # wait for the merkleblock command
            mb = self.wait_for(MerkleBlock)
            # check that the merkle block's hash is the same as the block hash
            if mb.hash() != block_hash:
                raise RuntimeError("Wrong block sent")
            # check that the merkle block is valid
            if not mb.is_valid():
                raise RuntimeError("Merkle Proof is invalid")
            # loop through the proved transactions from the Merkle block
            for tx_hash in mb.proved_txs():
                # wait for the tx command
                tx_obj = self.wait_for(Tx)
                # check that the hash matches
                if tx_obj.hash() != tx_hash:
                    raise RuntimeError(
                        "Wrong tx sent {} vs {}".format(tx_hash.hex(), tx_obj.id())
                    )
                # add to the results
                results.append(tx_obj)
        # return the results
        return results

    def is_tx_accepted(self, tx_obj):
        """Returns whether a transaction has been accepted on the network"""
        # sleep for a second to let everything propagate
        sleep(1)
        # create a GetDataMessage
        get_data = GetDataMessage()
        # ask for the tx
        get_data.add_data(TX_DATA_TYPE, tx_obj.hash())
        # send the GetDataMessage
        self.send(get_data)
        # now wait for a response
        got_tx = self.wait_for(Tx)
        if got_tx.id() == tx_obj.id():
            return True



from unittest import TestCase

from buidl.mnemonic import secure_mnemonic
from buidl.shamir import Share, ShareSet


class ShamirTest(TestCase):
    def test_share_errors(self):
        test_cases = [
            [
                "2. Mnemonic with invalid checksum (128 bits)",
                [
                    "duckling enlarge academic academic agency result length solution fridge kidney coal piece deal husband erode duke ajar critical decision kidney"
                ],
                ValueError,
            ],
            [
                "3. Mnemonic with invalid padding (128 bits)",
                [
                    "duckling enlarge academic academic email result length solution fridge kidney coal piece deal husband erode duke ajar music cargo fitness"
                ],
                SyntaxError,
            ],
            [
                "5. Basic sharing 2-of-3 (128 bits)",
                [
                    "shadow pistol academic always adequate wildlife fancy gross oasis cylinder mustang wrist rescue view short owner flip making coding armed"
                ],
                ValueError,
            ],
            [
                "6. Mnemonics with different identifiers (128 bits)",
                [
                    "adequate smoking academic acid debut wine petition glen cluster slow rhyme slow simple epidemic rumor junk tracks treat olympic tolerate",
                    "adequate stay academic agency agency formal party ting frequent learn upstairs remember smear leaf damage anatomy ladle market hush corner",
                ],
                TypeError,
            ],
            [
                "7. Mnemonics with different iteration exponents (128 bits)",
                [
                    "peasant leaves academic acid desert exact olympic math alive axle trial tackle drug deny decent smear dominant desert bucket remind",
                    "peasant leader academic agency cultural blessing percent network envelope medal junk primary human pumps jacket fragment payroll ticket evoke voice",
                ],
                TypeError,
            ],
            [
                "8. Mnemonics with mismatching group thresholds (128 bits)",
                [
                    "liberty category beard echo animal fawn temple briefing math username various wolf aviation fancy visual holy thunder yelp helpful payment",
                    "liberty category beard email beyond should fancy romp founder easel pink holy hairy romp loyalty material victim owner toxic custody",
                    "liberty category academic easy being hazard crush diminish oral lizard reaction cluster force dilemma deploy force club veteran expect photo",
                ],
                ValueError,
            ],
            [
                "9. Mnemonics with mismatching group counts (128 bits)",
                [
                    "average senior academic leaf broken teacher expect surface hour capture obesity desire negative dynamic dominant pistol mineral mailman iris aide",
                    "average senior academic agency curious pants blimp spew clothes slice script dress wrap firm shaft regular slavery negative theater roster",
                ],
                ValueError,
            ],
            [
                "10. Mnemonics with greater group threshold than group counts (128 bits)",
                [
                    "music husband acrobat acid artist finance center either graduate swimming object bike medical clothes station aspect spider maiden bulb welcome",
                    "music husband acrobat agency advance hunting bike corner density careful material civil evil tactics remind hawk discuss hobo voice rainbow",
                    "music husband beard academic black tricycle clock mayor estimate level photo episode exclude ecology papa source amazing salt verify divorce",
                ],
                ValueError,
            ],
            [
                "11. Mnemonics with duplicate member indices (128 bits)",
                [
                    "device stay academic always dive coal antenna adult black exceed stadium herald advance soldier busy dryer daughter evaluate minister laser",
                    "device stay academic always dwarf afraid robin gravity crunch adjust soul branch walnut coastal dream costume scholar mortgage mountain pumps",
                ],
                ValueError,
            ],
            [
                "12. Mnemonics with mismatching member thresholds (128 bits)",
                [
                    "hour painting academic academic device formal evoke guitar random modern justice filter withdraw trouble identify mailman insect general cover oven",
                    "hour painting academic agency artist again daisy capital beaver fiber much enjoy suitable symbolic identify photo editor romp float echo",
                ],
                ValueError,
            ],
            [
                "13. Mnemonics giving an invalid digest (128 bits)",
                [
                    "guilt walnut academic acid deliver remove equip listen vampire tactics nylon rhythm failure husband fatigue alive blind enemy teaspoon rebound",
                    "guilt walnut academic agency brave hamster hobo declare herd taste alpha slim criminal mild arcade formal romp branch pink ambition",
                ],
                ValueError,
            ],
            [
                "14. Insufficient number of groups (128 bits, case 1)",
                [
                    "eraser senior beard romp adorn nuclear spill corner cradle style ancient family general leader ambition exchange unusual garlic promise voice"
                ],
                ValueError,
            ],
            [
                "15. Insufficient number of groups (128 bits, case 2)",
                [
                    "eraser senior decision scared cargo theory device idea deliver modify curly include pancake both news skin realize vitamins away join",
                    "eraser senior decision roster beard treat identify grumpy salt index fake aviation theater cubic bike cause research dragon emphasis counter",
                ],
                ValueError,
            ],
            [
                "16. Threshold number of groups, but insufficient number of members in one group (128 bits)",
                [
                    "eraser senior decision shadow artist work morning estate greatest pipeline plan ting petition forget hormone flexible general goat admit surface",
                    "eraser senior beard romp adorn nuclear spill corner cradle style ancient family general leader ambition exchange unusual garlic promise voice",
                ],
                ValueError,
            ],
            [
                "21. Mnemonic with invalid checksum (256 bits)",
                [
                    "theory painting academic academic armed sweater year military elder discuss acne wildlife boring employer fused large satoshi bundle carbon diagnose anatomy hamster leaves tracks paces beyond phantom capital marvel lips brave detect lunar"
                ],
                ValueError,
            ],
            [
                "22. Mnemonic with invalid padding (256 bits)",
                [
                    "theory painting academic academic campus sweater year military elder discuss acne wildlife boring employer fused large satoshi bundle carbon diagnose anatomy hamster leaves tracks paces beyond phantom capital marvel lips facility obtain sister"
                ],
                SyntaxError,
            ],
            [
                "24. Basic sharing 2-of-3 (256 bits)",
                [
                    "humidity disease academic always aluminum jewelry energy woman receiver strategy amuse duckling lying evidence network walnut tactics forget hairy rebound impulse brother survive clothes stadium mailman rival ocean reward venture always armed unwrap"
                ],
                ValueError,
            ],
            [
                "25. Mnemonics with different identifiers (256 bits)",
                [
                    "smear husband academic acid deadline scene venture distance dive overall parking bracelet elevator justice echo burning oven chest duke nylon",
                    "smear isolate academic agency alpha mandate decorate burden recover guard exercise fatal force syndrome fumes thank guest drift dramatic mule",
                ],
                TypeError,
            ],
            [
                "26. Mnemonics with different iteration exponents (256 bits)",
                [
                    "finger trash academic acid average priority dish revenue academic hospital spirit western ocean fact calcium syndrome greatest plan losing dictate",
                    "finger traffic academic agency building lilac deny paces subject threaten diploma eclipse window unknown health slim piece dragon focus smirk",
                ],
                TypeError,
            ],
            [
                "27. Mnemonics with mismatching group thresholds (256 bits)",
                [
                    "flavor pink beard echo depart forbid retreat become frost helpful juice unwrap reunion credit math burning spine black capital lair",
                    "flavor pink beard email diet teaspoon freshman identify document rebound cricket prune headset loyalty smell emission skin often square rebound",
                    "flavor pink academic easy credit cage raisin crazy closet lobe mobile become drink human tactics valuable hand capture sympathy finger",
                ],
                ValueError,
            ],
            [
                "28. Mnemonics with mismatching group counts (256 bits)",
                [
                    "column flea academic leaf debut extra surface slow timber husky lawsuit game behavior husky swimming already paper episode tricycle scroll",
                    "column flea academic agency blessing garbage party software stadium verify silent umbrella therapy decorate chemical erode dramatic eclipse replace apart",
                ],
                ValueError,
            ],
            [
                "29. Mnemonics with greater group threshold than group counts (256 bits)",
                [
                    "smirk pink acrobat acid auction wireless impulse spine sprinkle fortune clogs elbow guest hush loyalty crush dictate tracks airport talent",
                    "smirk pink acrobat agency dwarf emperor ajar organize legs slice harvest plastic dynamic style mobile float bulb health coding credit",
                    "smirk pink beard academic alto strategy carve shame language rapids ruin smart location spray training acquire eraser endorse submit peaceful",
                ],
                ValueError,
            ],
            [
                "30. Mnemonics with duplicate member indices (256 bits)",
                [
                    "fishing recover academic always device craft trend snapshot gums skin downtown watch device sniff hour clock public maximum garlic born",
                    "fishing recover academic always aircraft view software cradle fangs amazing package plastic evaluate intend penalty epidemic anatomy quarter cage apart",
                ],
                ValueError,
            ],
            [
                "31. Mnemonics with mismatching member thresholds (256 bits)",
                [
                    "evoke garden academic academic answer wolf scandal modern warmth station devote emerald market physics surface formal amazing aquatic gesture medical",
                    "evoke garden academic agency deal revenue knit reunion decrease magazine flexible company goat repair alarm military facility clogs aide mandate",
                ],
                ValueError,
            ],
            [
                "32. Mnemonics giving an invalid digest (256 bits)",
                [
                    "river deal academic acid average forbid pistol peanut custody bike class aunt hairy merit valid flexible learn ajar very easel",
                    "river deal academic agency camera amuse lungs numb isolate display smear piece traffic worthy year patrol crush fact fancy emission",
                ],
                ValueError,
            ],
            [
                "33. Insufficient number of groups (256 bits, case 1)",
                [
                    "wildlife deal beard romp alcohol space mild usual clothes union nuclear testify course research heat listen task location thank hospital slice smell failure fawn helpful priest ambition average recover lecture process dough stadium"
                ],
                ValueError,
            ],
            [
                "34. Insufficient number of groups (256 bits, case 2)",
                [
                    "wildlife deal decision scared acne fatal snake paces obtain election dryer dominant romp tactics railroad marvel trust helpful flip peanut theory theater photo luck install entrance taxi step oven network dictate intimate listen",
                    "wildlife deal decision smug ancestor genuine move huge cubic strategy smell game costume extend swimming false desire fake traffic vegan senior twice timber submit leader payroll fraction apart exact forward pulse tidy install",
                ],
                ValueError,
            ],
            [
                "35. Threshold number of groups, but insufficient number of members in one group (256 bits)",
                [
                    "wildlife deal decision shadow analysis adjust bulb skunk muscle mandate obesity total guitar coal gravity carve slim jacket ruin rebuild ancestor numerous hour mortgage require herd maiden public ceiling pecan pickup shadow club",
                    "wildlife deal beard romp alcohol space mild usual clothes union nuclear testify course research heat listen task location thank hospital slice smell failure fawn helpful priest ambition average recover lecture process dough stadium",
                ],
                ValueError,
            ],
            [
                "39. Mnemonic with insufficient length",
                [
                    "junk necklace academic academic acne isolate join hesitate lunar roster dough calcium chemical ladybug amount mobile glasses verify cylinder"
                ],
                ValueError,
            ],
            [
                "40. Mnemonic with invalid master secret length",
                [
                    "fraction necklace academic academic award teammate mouse regular testify coding building member verdict purchase blind camera duration email prepare spirit quarter"
                ],
                SyntaxError,
            ],
        ]
        for test_name, mnemonics, error in test_cases:
            with self.assertRaises(error):
                share_set = ShareSet([Share.parse(m) for m in mnemonics])
                share_set.recover(b"TREZOR")

    def test_recover(self):
        test_cases = [
            [
                "1. Valid mnemonic without sharing (128 bits)",
                [
                    "duckling enlarge academic academic agency result length solution fridge kidney coal piece deal husband erode duke ajar critical decision keyboard"
                ],
                "bb54aac4b89dc868ba37d9cc21b2cece",
            ],
            [
                "4. Basic sharing 2-of-3 (128 bits)",
                [
                    "shadow pistol academic always adequate wildlife fancy gross oasis cylinder mustang wrist rescue view short owner flip making coding armed",
                    "shadow pistol academic acid actress prayer class unknown daughter sweater depict flip twice unkind craft early superior advocate guest smoking",
                ],
                "b43ceb7e57a0ea8766221624d01b0864",
            ],
            [
                "17. Threshold number of groups and members in each group (128 bits, case 1)",
                [
                    "eraser senior decision roster beard treat identify grumpy salt index fake aviation theater cubic bike cause research dragon emphasis counter",
                    "eraser senior ceramic snake clay various huge numb argue hesitate auction category timber browser greatest hanger petition script leaf pickup",
                    "eraser senior ceramic shaft dynamic become junior wrist silver peasant force math alto coal amazing segment yelp velvet image paces",
                    "eraser senior ceramic round column hawk trust auction smug shame alive greatest sheriff living perfect corner chest sled fumes adequate",
                    "eraser senior decision smug corner ruin rescue cubic angel tackle skin skunk program roster trash rumor slush angel flea amazing",
                ],
                "7c3397a292a5941682d7a4ae2d898d11",
            ],
            [
                "18. Threshold number of groups and members in each group (128 bits, case 2)",
                [
                    "eraser senior decision smug corner ruin rescue cubic angel tackle skin skunk program roster trash rumor slush angel flea amazing",
                    "eraser senior beard romp adorn nuclear spill corner cradle style ancient family general leader ambition exchange unusual garlic promise voice",
                    "eraser senior decision scared cargo theory device idea deliver modify curly include pancake both news skin realize vitamins away join",
                ],
                "7c3397a292a5941682d7a4ae2d898d11",
            ],
            [
                "19. Threshold number of groups and members in each group (128 bits, case 3)",
                [
                    "eraser senior beard romp adorn nuclear spill corner cradle style ancient family general leader ambition exchange unusual garlic promise voice",
                    "eraser senior acrobat romp bishop medical gesture pumps secret alive ultimate quarter priest subject class dictate spew material endless market",
                ],
                "7c3397a292a5941682d7a4ae2d898d11",
            ],
            [
                "20. Valid mnemonic without sharing (256 bits)",
                [
                    "theory painting academic academic armed sweater year military elder discuss acne wildlife boring employer fused large satoshi bundle carbon diagnose anatomy hamster leaves tracks paces beyond phantom capital marvel lips brave detect luck"
                ],
                "989baf9dcaad5b10ca33dfd8cc75e42477025dce88ae83e75a230086a0e00e92",
            ],
            [
                "23. Basic sharing 2-of-3 (256 bits)",
                [
                    "humidity disease academic always aluminum jewelry energy woman receiver strategy amuse duckling lying evidence network walnut tactics forget hairy rebound impulse brother survive clothes stadium mailman rival ocean reward venture always armed unwrap",
                    "humidity disease academic agency actress jacket gross physics cylinder solution fake mortgage benefit public busy prepare sharp friar change work slow purchase ruler again tricycle involve viral wireless mixture anatomy desert cargo upgrade",
                ],
                "c938b319067687e990e05e0da0ecce1278f75ff58d9853f19dcaeed5de104aae",
            ],
            [
                "36. Threshold number of groups and members in each group (256 bits, case 1)",
                [
                    "wildlife deal ceramic round aluminum pitch goat racism employer miracle percent math decision episode dramatic editor lily prospect program scene rebuild display sympathy have single mustang junction relate often chemical society wits estate",
                    "wildlife deal decision scared acne fatal snake paces obtain election dryer dominant romp tactics railroad marvel trust helpful flip peanut theory theater photo luck install entrance taxi step oven network dictate intimate listen",
                    "wildlife deal ceramic scatter argue equip vampire together ruin reject literary rival distance aquatic agency teammate rebound false argue miracle stay again blessing peaceful unknown cover beard acid island language debris industry idle",
                    "wildlife deal ceramic snake agree voter main lecture axis kitchen physics arcade velvet spine idea scroll promise platform firm sharp patrol divorce ancestor fantasy forbid goat ajar believe swimming cowboy symbolic plastic spelling",
                    "wildlife deal decision shadow analysis adjust bulb skunk muscle mandate obesity total guitar coal gravity carve slim jacket ruin rebuild ancestor numerous hour mortgage require herd maiden public ceiling pecan pickup shadow club",
                ],
                "5385577c8cfc6c1a8aa0f7f10ecde0a3318493262591e78b8c14c6686167123b",
            ],
            [
                "37. Threshold number of groups and members in each group (256 bits, case 2)",
                [
                    "wildlife deal decision scared acne fatal snake paces obtain election dryer dominant romp tactics railroad marvel trust helpful flip peanut theory theater photo luck install entrance taxi step oven network dictate intimate listen",
                    "wildlife deal beard romp alcohol space mild usual clothes union nuclear testify course research heat listen task location thank hospital slice smell failure fawn helpful priest ambition average recover lecture process dough stadium",
                    "wildlife deal decision smug ancestor genuine move huge cubic strategy smell game costume extend swimming false desire fake traffic vegan senior twice timber submit leader payroll fraction apart exact forward pulse tidy install",
                ],
                "5385577c8cfc6c1a8aa0f7f10ecde0a3318493262591e78b8c14c6686167123b",
            ],
            [
                "38. Threshold number of groups and members in each group (256 bits, case 3)",
                [
                    "wildlife deal beard romp alcohol space mild usual clothes union nuclear testify course research heat listen task location thank hospital slice smell failure fawn helpful priest ambition average recover lecture process dough stadium",
                    "wildlife deal acrobat romp anxiety axis starting require metric flexible geology game drove editor edge screw helpful have huge holy making pitch unknown carve holiday numb glasses survive already tenant adapt goat fangs",
                ],
                "5385577c8cfc6c1a8aa0f7f10ecde0a3318493262591e78b8c14c6686167123b",
            ],
        ]
        for test_name, mnemonics, expected in test_cases:
            share_set = ShareSet([Share.parse(m) for m in mnemonics])
            self.assertEqual(share_set.recover(b"TREZOR").hex(), expected, test_name)

    def test_split(self):
        secret = bytes.fromhex("7c3397a292a5941682d7a4ae2d898d11")
        for k, n in ((2, 3), (3, 5), (5, 5), (9, 9), (13, 15)):
            share_data = ShareSet.split_secret(secret, k, n)
            self.assertEqual(secret, ShareSet.interpolate(255, share_data[:k]))

    def test_generate(self):
        for num_bits in (128, 256):
            bip39_mnemonic = secure_mnemonic(num_bits)
            passphrase = b"buidltest"
            for k, n in ((2, 3), (3, 5), (5, 5), (9, 9), (13, 15), (2, 8)):
                slip39_mnemonics = ShareSet.generate_shares(
                    bip39_mnemonic, k, n, passphrase=passphrase, exponent=2
                )
                self.assertEqual(
                    ShareSet.recover_mnemonic(slip39_mnemonics[:k], passphrase),
                    bip39_mnemonic,
                )
            slip39_mnemonics = ShareSet.generate_shares(
                bip39_mnemonic, 2, 7, passphrase=passphrase
            )
            for s1 in slip39_mnemonics:
                for s2 in set(slip39_mnemonics) - {s1}:
                    self.assertEqual(
                        ShareSet.recover_mnemonic([s1, s2], passphrase), bip39_mnemonic
                    )



from unittest import TestCase

from buidl.ecc import G, S256Point, PrivateKey, Signature
from buidl.bech32 import decode_bech32

from random import randint


class S256Test(TestCase):
    #    def test_order(self):
    #        point = N * G
    #        self.assertIsNone(point.x)

    def test_pubpoint(self):
        # write a test that tests the public point for the following
        points = (
            # secret, x, y
            (
                7,
                "045CBDF0646E5DB4EAA398F365F2EA7A0E3D419B7E0330E39CE92BDDEDCAC4F9BC6AEBCA40BA255960A3178D6D861A54DBA813D0B813FDE7B5A5082628087264DA",
            ),
            (
                1485,
                "04C982196A7466FBBBB0E27A940B6AF926C1A74D5AD07128C82824A11B5398AFDA7A91F9EAE64438AFB9CE6448A1C133DB2D8FB9254E4546B6F001637D50901F55",
            ),
            (
                2 ** 128,
                "048F68B9D2F63B5F339239C1AD981F162EE88C5678723EA3351B7B444C9EC4C0DA662A9F2DBA063986DE1D90C2B6BE215DBBEA2CFE95510BFDF23CBF79501FFF82",
            ),
            (
                2 ** 240 + 2 ** 31,
                "049577FF57C8234558F293DF502CA4F09CBC65A6572C842B39B366F2171794511610B49C67FA9365AD7B90DAB070BE339A1DAF9052373EC30FFAE4F72D5E66D053",
            ),
        )

        # iterate over points
        for secret, sec in points:
            # initialize the secp256k1 point (S256Point)
            point = S256Point.parse(bytes.fromhex(sec))
            # check that the secret*G is the same as the point
            self.assertEqual(secret * G, point)

    def test_sec(self):
        coefficient = 999 ** 3
        uncompressed = "049d5ca49670cbe4c3bfa84c96a8c87df086c6ea6a24ba6b809c9de234496808d56fa15cc7f3d38cda98dee2419f415b7513dde1301f8643cd9245aea7f3f911f9"
        compressed = (
            "039d5ca49670cbe4c3bfa84c96a8c87df086c6ea6a24ba6b809c9de234496808d5"
        )
        point = coefficient * G
        self.assertEqual(point.sec(compressed=False), bytes.fromhex(uncompressed))
        self.assertEqual(point.sec(compressed=True), bytes.fromhex(compressed))
        coefficient = 123
        uncompressed = "04a598a8030da6d86c6bc7f2f5144ea549d28211ea58faa70ebf4c1e665c1fe9b5204b5d6f84822c307e4b4a7140737aec23fc63b65b35f86a10026dbd2d864e6b"
        compressed = (
            "03a598a8030da6d86c6bc7f2f5144ea549d28211ea58faa70ebf4c1e665c1fe9b5"
        )
        point = coefficient * G
        self.assertEqual(point.sec(compressed=False), bytes.fromhex(uncompressed))
        self.assertEqual(point.sec(compressed=True), bytes.fromhex(compressed))
        coefficient = 42424242
        uncompressed = "04aee2e7d843f7430097859e2bc603abcc3274ff8169c1a469fee0f20614066f8e21ec53f40efac47ac1c5211b2123527e0e9b57ede790c4da1e72c91fb7da54a3"
        compressed = (
            "03aee2e7d843f7430097859e2bc603abcc3274ff8169c1a469fee0f20614066f8e"
        )
        point = coefficient * G
        self.assertEqual(point.sec(compressed=False), bytes.fromhex(uncompressed))
        self.assertEqual(point.sec(compressed=True), bytes.fromhex(compressed))

    def test_address(self):
        tests = (
            (
                888 ** 3,
                "148dY81A9BmdpMhvYEVznrM45kWN32vSCN",
                "mnabU9NCcRE5zcNZ2C16CnvKPELrFvisn3",
            ),
            (
                321,
                "1FNgueDbMYjNQ8HT77sHKxTwdrHMdTGwyN",
                "mfx3y63A7TfTtXKkv7Y6QzsPFY6QCBCXiP",
            ),
            (
                4242424242,
                "1HUYfVCXEmp76uh17bE2gA72Vuqv4wrM1a",
                "mgY3bVusRUL6ZB2Ss999CSrGVbdRwVpM8s",
            ),
        )
        for secret, mainnet_legacy, testnet_legacy in tests:
            point = secret * G
            self.assertEqual(point.address(network="mainnet"), mainnet_legacy)
            self.assertEqual(
                point.address(compressed=False, network="testnet"), testnet_legacy
            )
            self.assertEqual(
                point.address(compressed=False, network="signet"), testnet_legacy
            )

    def test_bech32_address(self):
        tests = (
            (
                888 ** 3,
                "bc1qyfvunnpszmjwcqgfk9dsne6j4edq3fglx9y5x7",
                "tb1qyfvunnpszmjwcqgfk9dsne6j4edq3fglvrl8ad",
            ),
            (
                321,
                "bc1qnk4u7vkat6ck9t4unlgvvle8dhsqp40mrssamm",
                "tb1qnk4u7vkat6ck9t4unlgvvle8dhsqp40mfktwqg",
            ),
            (
                4242424242,
                "bc1qkjm6e3c79zy7clsfx86q4pvy46ccc5u9xa6f6e",
                "tb1qkjm6e3c79zy7clsfx86q4pvy46ccc5u9vmp6p2",
            ),
        )
        for secret, mainnet_bech32, testnet_bech32 in tests:
            point = secret * G
            self.assertEqual(point.bech32_address(network="mainnet"), mainnet_bech32)
            self.assertEqual(decode_bech32(mainnet_bech32)[2], point.hash160())
            self.assertEqual(point.bech32_address(network="testnet"), testnet_bech32)
            self.assertEqual(point.bech32_address(network="signet"), testnet_bech32)

    def test_p2sh_p2wpkh_address(self):
        tests = (
            (
                888 ** 3,
                "32cE3VHX5k1Z4gDCJBXMSLgd1akUzvqNvH",
                "2MtAS7EDYhCWuGTqjyK9E4HftDvxek7ELQn",
            ),
            (
                321,
                "3KPpFmmGNoKi5ikrH4QsMNmNnQtzkdw4Kx",
                "2NAx2KWhHzFq4HWPPxC2jyKkdzm7AVsEge4",
            ),
            (
                4242424242,
                "3M7oCrExZ6ZYjyn2oxXxYnE14m813espco",
                "2NCg1GbAzAZ4twmQaV69qAjDGH7LApz5kA4",
            ),
        )
        for secret, mainnet_p2sh, testnet_p2sh in tests:
            point = secret * G
            self.assertEqual(point.p2sh_p2wpkh_address(network="mainnet"), mainnet_p2sh)
            self.assertEqual(point.p2sh_p2wpkh_address(network="testnet"), testnet_p2sh)
            self.assertEqual(point.p2sh_p2wpkh_address(network="signet"), testnet_p2sh)

    def test_verify(self):
        tests = (
            (
                0xEC208BAA0FC1C19F708A9CA96FDEFF3AC3F230BB4A7BA4AEDE4942AD003C0F60,
                "3045022100ac8d1c87e51d0d441be8b3dd5b05c8795b48875dffe00b7ffcfac23010d3a3950220068342ceff8935ededd102dd876ffd6ba72d6a427a3edb13d26eb0781cb423c4",
                "04887387e452b8eacc4acfde10d9aaf7f6d9a0f975aabb10d006e4da568744d06c61de6d95231cd89026e286df3b6ae4a894a3378e393e93a0f45b666329a0ae34",
            ),
            (
                0x7C076FF316692A3D7EB3C3BB0F8B1488CF72E1AFCD929E29307032997A838A3D,
                "3044022000eff69ef2b1bd93a66ed5219add4fb51e11a840f404876325a1e8ffe0529a2c022038df8011e682d839e75159debf909408cb3f12ae472b1d88cf6280cf01c6568b",
                "04887387e452b8eacc4acfde10d9aaf7f6d9a0f975aabb10d006e4da568744d06c61de6d95231cd89026e286df3b6ae4a894a3378e393e93a0f45b666329a0ae34",
            ),
            (
                0x2270CB0316E68389A3A23DE16023A03B8FC271A21B467B1DC97E0FC0E2CE97F7,
                "3045022100ea6d640d5275d091607e1f4ad5cdb214e45f8d17cca1095074894dde347605ba022029062e1ff0d9eee52da1f3621caf92436877d7076720e2b3d9f226bf853e2b75",
                "04f47dc2ac0ecaadda5ee2b3ab9bc4e02c3eafb2abcc426643686ad95f6d4e8c44e33fa47d96fc2dace0ef2f583965cf6a0f8faa7a070c0f8ee986d192e2d21835",
            ),
        )
        for z, der_hex, sec in tests:
            point = S256Point.parse(bytes.fromhex(sec))
            der = bytes.fromhex(der_hex)
            self.assertTrue(point.verify(z, Signature.parse(der)))

    def test_parse(self):
        csec = bytes.fromhex(
            "0349fc4e631e3624a545de3f89f5d8684c7b8138bd94bdd531d2e213bf016b278a"
        )
        point = S256Point.parse(csec)
        usec = bytes.fromhex(
            "0449fc4e631e3624a545de3f89f5d8684c7b8138bd94bdd531d2e213bf016b278aa56c896489c71dfc65701ce25050f542f336893fb8cd15f4e8e5c124dbf58e47"
        )
        self.assertEqual(point.sec(False), usec)


class SignatureTest(TestCase):
    def test_der(self):
        tests = (
            "3045022100ed81ff192e75a3fd2304004dcadb746fa5e24c5031ccfcf21320b0277457c98f02207a986d955c6e0cb35d446a89d3f56100f4d7f67801c31967743a9c8e10615bed",
        )
        for der_hex in tests:
            der = bytes.fromhex(der_hex)
            sig = Signature.parse(der)
            self.assertTrue("Signature" in sig.__repr__())
            computed = sig.der()
            self.assertEqual(der, computed)


class PrivateKeyTest(TestCase):
    def test_sign(self):
        pk = PrivateKey(randint(0, 2 ** 256))
        z = randint(0, 2 ** 256)
        sig = pk.sign(z)
        self.assertTrue(pk.point.verify(z, sig))

    def test_sign_message(self):
        pk = PrivateKey(randint(0, 2 ** 256))
        message = b"This is a test message"
        sig = pk.sign_message(message)
        self.assertTrue(pk.point.verify_message(message, sig))



from io import BytesIO

from buidl.ecc import N, PrivateKey, S256Point
from buidl.helper import (
    big_endian_to_int,
    byte_to_int,
    encode_base58_checksum,
    hmac_sha512,
    hmac_sha512_kdf,
    int_to_big_endian,
    int_to_byte,
    is_intable,
    raw_decode_base58,
)
from buidl.mnemonic import (
    BIP39,
    InvalidChecksumWordsError,
    secure_mnemonic,
    mnemonic_to_bytes,
)
from buidl.shamir import ShareSet

# https://github.com/satoshilabs/slips/blob/master/slip-0132.md
# Make default xpub/xpriv, but allow for other version bytes

XPRV = {
    "mainnet": bytes.fromhex("0488ade4"),
    "testnet": bytes.fromhex("04358394"),
    "signet": bytes.fromhex("04358394"),
}

XPUB = {
    "mainnet": bytes.fromhex("0488b21e"),
    "testnet": bytes.fromhex("043587cf"),
    "signet": bytes.fromhex("043587cf"),
}

# P2PKH or P2SH, P2WPKH in P2SH, P2WPKH, Multi-signature P2WSH in P2SH, Multi-signature P2WSH
ALL_MAINNET_XPRVS = {
    bytes.fromhex(x)
    for x in ["0488ade4", "049d7878", "04b2430c", "0295b005", "02aa7a99"]
}
ALL_MAINNET_XPUBS = {
    bytes.fromhex(x)
    for x in ["0488b21e", "049d7cb2", "04b24746", "0295b43f", "02aa7ed3"]
}
ALL_TESTNET_XPRVS = {
    bytes.fromhex(x)
    for x in ["04358394", "044a4e28", "045f18bc", "024285b5", "02575048"]
}
ALL_TESTNET_XPUBS = {
    bytes.fromhex(x)
    for x in ["043587cf", "044a5262", "045f1cf6", "024289ef", "02575483"]
}

DEFAULT_P2WSH_PATH = {
    "mainnet": "m/48h/0h/0h/2h",
    "testnet": "m/48h/1h/0h/2h",
    "signet": "m/48h/1h/0h/2h",
}


class HDPrivateKey:
    def __init__(
        self,
        private_key,
        chain_code,
        depth=0,
        parent_fingerprint=b"\x00\x00\x00\x00",
        child_number=0,
        network="mainnet",
        priv_version=None,
        pub_version=None,
    ):
        # the main secret, should be a PrivateKey object
        self.private_key = private_key
        self.private_key.network = network
        # the code to make derivation deterministic
        self.chain_code = chain_code
        # level the current key is at in the heirarchy
        self.depth = depth
        # fingerprint of the parent key
        self.parent_fingerprint = parent_fingerprint
        # what order child this is
        self.child_number = child_number
        self.network = network

        if priv_version is None:
            # Set priv_version based on whether or not we are testnet
            priv_version = XPRV[self.network]
        self.priv_version = priv_version

        # keep a copy of the corresponding public key
        self.pub = HDPublicKey(
            point=private_key.point,
            chain_code=chain_code,
            depth=depth,
            parent_fingerprint=parent_fingerprint,
            child_number=child_number,
            network=network,
            pub_version=pub_version,  # HDPublicKey handles the case where pub_version is None
        )

    def wif(self):
        return self.private_key.wif()

    def sec(self):
        return self.pub.sec()

    def hash160(self):
        return self.pub.hash160()

    def p2pkh_script(self):
        return self.pub.p2pkh_script()

    def p2wpkh_script(self):
        return self.pub.p2wpkh_script()

    def p2sh_p2wpkh_script(self):
        return self.pub.p2sh_p2wpkh_script()

    def address(self):
        return self.pub.address()

    def bech32_address(self):
        return self.pub.bech32_address()

    def p2sh_p2wpkh_address(self):
        return self.pub.p2sh_p2wpkh_address()

    def __repr__(self):
        return self.xprv()

    @classmethod
    def from_seed(cls, seed, network="mainnet", priv_version=None, pub_version=None):
        # get hmac_sha512 with b'Bitcoin seed' and seed
        h = hmac_sha512(b"Bitcoin seed", seed)
        # create the private key using the first 32 bytes in big endian
        private_key = PrivateKey(secret=big_endian_to_int(h[:32]))
        # chaincode is the last 32 bytes
        chain_code = h[32:]
        # return an instance of the class
        return cls(
            private_key=private_key,
            chain_code=chain_code,
            network=network,
            priv_version=priv_version,
            pub_version=pub_version,
        )

    def child(self, index):
        """Returns the child HDPrivateKey at a particular index.
        Hardened children return for indices >= 0x8000000.
        """
        # if index >= 0x80000000
        if index >= 0x80000000:
            # the message data is the private key secret in 33 bytes in
            #  big-endian and the index in 4 bytes big-endian.
            data = int_to_big_endian(self.private_key.secret, 33) + int_to_big_endian(
                index, 4
            )
        else:
            # the message data is the public key compressed SEC
            #  and the index in 4 bytes big-endian.
            data = self.private_key.point.sec() + int_to_big_endian(index, 4)
        # get the hmac_sha512 with chain code and data
        h = hmac_sha512(self.chain_code, data)
        # the new secret is the first 32 bytes as a big-endian integer
        #  plus the secret mod N
        secret = (big_endian_to_int(h[:32]) + self.private_key.secret) % N
        # create the PrivateKey object
        private_key = PrivateKey(secret=secret)
        # the chain code is the last 32 bytes
        chain_code = h[32:]
        # depth is whatever the current depth + 1
        depth = self.depth + 1
        # parent_fingerprint is the fingerprint of this node
        parent_fingerprint = self.fingerprint()
        # child number is the index
        child_number = index
        # return a new HDPrivateKey instance
        return HDPrivateKey(
            private_key=private_key,
            chain_code=chain_code,
            depth=depth,
            parent_fingerprint=parent_fingerprint,
            child_number=child_number,
            network=self.network,
            priv_version=self.priv_version,
            pub_version=self.pub.pub_version,
        )

    def traverse(self, path):
        """Returns the HDPrivateKey at the path indicated.
        Path should be in the form of m/x/y/z where x' means
        hardened"""

        # accept path in uppercase and/or using h instead of '
        path = path.lower().replace("h", "'")

        if not path.startswith("m"):
            raise ValueError(f"Invalid Path: {path}")

        # keep track of the current node starting with self
        current = self
        # split up the path by the '/' splitter, ignore the first
        components = path.split("/")[1:]
        # iterate through the path components
        for child in components:
            # if the child ends with a ', we have a hardened child
            if child.endswith("'"):
                # index is the integer representation + 0x80000000
                index = int(child[:-1]) + 0x80000000
            # else the index is the integer representation
            else:
                index = int(child)
            # grab the child at the index calculated
            current = current.child(index)
        # return the current child
        return current

    def raw_serialize(self, priv_version):
        # version + depth + parent_fingerprint + child number + chain code + private key
        # start with priv_version, which should be a constant
        raw = priv_version
        # add depth, which is 1 byte using int_to_byte
        raw += int_to_byte(self.depth)
        # add the parent_fingerprint
        raw += self.parent_fingerprint
        # add the child number 4 bytes using int_to_big_endian
        raw += int_to_big_endian(self.child_number, 4)
        # add the chain code
        raw += self.chain_code
        # add the 0 byte and the private key's secret in big endian, 33 bytes
        raw += int_to_big_endian(self.private_key.secret, 33)
        return raw

    def xprv(self, version=None):
        """Returns the base58-encoded x/y/z prv."""
        if version is None:
            version = self.priv_version
        raw = self.raw_serialize(version)
        return encode_base58_checksum(raw)

    def xpub(self, version=None):
        return self.pub.xpub(version=version)

    # passthrough methods
    def fingerprint(self):
        return self.pub.fingerprint()

    @classmethod
    def parse(cls, s):
        """Returns a HDPrivateKey from an extended key string"""
        # get the bytes from the base58 using raw_decode_base58
        raw = raw_decode_base58(s)
        # check that the length of the raw is 78 bytes, otherwise raise ValueError
        if len(raw) != 78:
            raise ValueError("Not a proper extended key")
        # create a stream
        stream = BytesIO(raw)
        # return the raw parsing of the stream
        return cls.raw_parse(stream)

    @classmethod
    def raw_parse(cls, s):
        """Returns a HDPrivateKey from a stream"""
        # first 4 bytes are the priv_version
        priv_version = s.read(4)
        # check that the priv_version is one of the TESTNET or MAINNET
        #  private keys, if not raise a ValueError
        if priv_version in ALL_TESTNET_XPRVS:
            network = "testnet"
        elif priv_version in ALL_MAINNET_XPRVS:
            network = "mainnet"
        else:
            raise ValueError(f"not a valid [t-z]prv: {priv_version}")
        # the next byte is depth
        depth = byte_to_int(s.read(1))
        # next 4 bytes are the parent_fingerprint
        parent_fingerprint = s.read(4)
        # next 4 bytes is the child number in big-endian
        child_number = big_endian_to_int(s.read(4))
        # next 32 bytes are the chain code
        chain_code = s.read(32)
        # the next byte should be b'\x00'
        if byte_to_int(s.read(1)) != 0:
            raise ValueError("private key should be preceded by a zero byte")
        # last 32 bytes should be the private key in big endian
        private_key = PrivateKey(secret=big_endian_to_int(s.read(32)))
        # return an instance of the class
        return cls(
            private_key=private_key,
            chain_code=chain_code,
            depth=depth,
            parent_fingerprint=parent_fingerprint,
            child_number=child_number,
            network=network,
            priv_version=priv_version,
            # HDPublicKey will handle its own versioning:
            pub_version=None,
        )

    def _get_address(self, purpose, account=0, external=True, address=0):
        """Returns the proper address among purposes 44', 49' and 84'.
        p2pkh for 44', p2sh-p2wpkh for 49' and p2wpkh for 84'."""
        # if purpose is not one of 44', 49' or 84', raise ValueError
        if purpose not in ("44'", "49'", "84'"):
            raise ValueError(
                "Cannot create an address without a proper purpose: {}".format(purpose)
            )
        # if testnet/signet, coin is 1', otherwise 0'
        if self.network == "mainnet":
            coin = "0'"
        else:
            coin = "1'"
        # if external, chain is 0, otherwise 1
        if external:
            chain = "0"
        else:
            chain = "1"
        # create the path m/purpose'/coin'/account'/chain/address
        path = "m/{}/{}/{}'/{}/{}".format(purpose, coin, account, chain, address)
        # get the HDPrivateKey at that location
        hd_priv = self.traverse(path)
        # if 44', return the address
        if purpose == "44'":
            return hd_priv.address()
        # if 49', return the p2sh_p2wpkh_address
        elif purpose == "49'":
            return hd_priv.p2sh_p2wpkh_address()
        # if 84', return the bech32_address
        elif purpose == "84'":
            return hd_priv.bech32_address()

    def get_p2pkh_receiving_address(self, account=0, address=0):
        return self._get_address("44'", account, True, address)

    def get_p2pkh_change_address(self, account=0, address=0):
        return self._get_address("44'", account, False, address)

    def get_p2sh_p2wpkh_receiving_address(self, account=0, address=0):
        return self._get_address("49'", account, True, address)

    def get_p2sh_p2wpkh_change_address(self, account=0, address=0):
        return self._get_address("49'", account, False, address)

    def get_p2wpkh_receiving_address(self, account=0, address=0):
        return self._get_address("84'", account, True, address)

    def get_p2wpkh_change_address(self, account=0, address=0):
        return self._get_address("84'", account, False, address)

    @classmethod
    def generate(
        cls,
        password=b"",
        extra_entropy=0,
        network="mainnet",
        priv_version=None,
        pub_version=None,
    ):
        mnemonic = secure_mnemonic(extra_entropy=extra_entropy)
        return mnemonic, cls.from_mnemonic(
            mnemonic,
            password=password,
            network=network,
            priv_version=priv_version,
            pub_version=pub_version,
        )

    @classmethod
    def from_mnemonic(
        cls,
        mnemonic,
        password=b"",
        path="m",
        network="mainnet",
        priv_version=None,
        pub_version=None,
    ):
        """Returns a HDPrivateKey object from the mnemonic."""
        # this will check that the mnemonic is valid
        mnemonic_to_bytes(mnemonic)
        # normalize in case we got a mnemonic that's just the first 4 letters
        normalized = " ".join([BIP39.normalize(word) for word in mnemonic.split()])
        # salt is b'mnemonic' + password
        salt = b"mnemonic" + password
        # the seed is the hmac_sha512_kdf with normalized mnemonic and salt
        seed = hmac_sha512_kdf(normalized, salt)
        # return the HDPrivateKey at the path specified
        return cls.from_seed(
            seed, network=network, priv_version=priv_version, pub_version=pub_version
        ).traverse(path)

    @classmethod
    def from_shares(
        cls, share_mnemonics, passphrase=b"", password=b"", path="m", network="mainnet"
    ):
        """Returns a HDPrivateKey object from SLIP39 mnemonics.
        Note passphrase is for the shares and
        password is for the mnemonic."""
        mnemonic = ShareSet.recover_mnemonic(share_mnemonics, passphrase)
        return cls.from_mnemonic(mnemonic, password, path, network)

    def generate_p2wsh_key_record(
        self, bip32_path=None, use_slip132_version_byte=False
    ):
        """
        Convenience method for generating a public key_record to supply to your Coordinator software.
        """
        # Check that we're using the root HDPrivateKey
        if self.depth != 0:
            raise ValueError(
                "Key depth != 0. Please supply the root HDPrivateKey to use this method."
            )
        if self.parent_fingerprint != b"\x00\x00\x00\x00":
            raise ValueError(
                "Parent fingerprint != the zero byte. Please supply the root HDPrivateKey to use this method."
            )
        if self.child_number != 0:
            raise ValueError(
                "child_number != 0. Please supply the root HDPrivateKey to use this method."
            )
        if bip32_path is not None and not is_valid_bip32_path(bip32_path):
            raise ValueError(f"bip32_path {bip32_path} is not valid")

        if use_slip132_version_byte:
            # https://github.com/satoshilabs/slips/blob/master/slip-0132.md
            if self.network == "mainnet":
                version_byte = bytes.fromhex("02aa7ed3")
            else:
                version_byte = bytes.fromhex("02575483")
        else:
            version_byte = None  # buidl will automatically pick between xpub/tpub

        if bip32_path is None:
            bip32_path = DEFAULT_P2WSH_PATH[self.network]
        else:
            # make the standard h instead of ' for consistency
            bip32_path = bip32_path.replace("'", "h")

        xpub = self.traverse(bip32_path).xpub(version=version_byte)
        return f"[{self.fingerprint().hex()}/{bip32_path[2:]}]{xpub}"


class HDPublicKey:
    def __init__(
        self,
        point,
        chain_code,
        depth,
        parent_fingerprint,
        child_number,
        network="mainnet",
        pub_version=None,
    ):
        self.point = point
        self.chain_code = chain_code
        self.depth = depth
        self.parent_fingerprint = parent_fingerprint
        self.child_number = child_number
        self.network = network
        if pub_version is None:
            # Set pub_version based on whether or not we are testnet
            pub_version = XPUB[network]
        self.pub_version = pub_version
        self._raw = None

    def __repr__(self):
        return self.xpub()

    def sec(self):
        return self.point.sec()

    def hash160(self):
        return self.point.hash160()

    def p2pkh_script(self):
        return self.point.p2pkh_script()

    def p2wpkh_script(self):
        return self.point.p2wpkh_script()

    def p2sh_p2wpkh_script(self):
        return self.point.p2sh_p2wpkh_script()

    def address(self):
        return self.point.address(network=self.network)

    def bech32_address(self):
        return self.point.bech32_address(network=self.network)

    def p2sh_p2wpkh_address(self):
        return self.point.p2sh_p2wpkh_address(network=self.network)

    def fingerprint(self):
        """Fingerprint is the hash160's first 4 bytes"""
        return self.hash160()[:4]

    def child(self, index):
        """Returns the child HDPrivateKey at a particular index.
        Raises ValueError for indices >= 0x8000000.
        """
        # if index >= 0x80000000, raise a ValueError
        if index >= 0x80000000:
            raise ValueError("child number should always be less than 2^31")
        # data is the SEC compressed and the index in 4 bytes big-endian
        data = self.point.sec() + int_to_big_endian(index, 4)
        # get hmac_sha512 with chain code, data
        h = hmac_sha512(self.chain_code, data)
        # the new public point is the current point +
        #  the first 32 bytes in big endian * G
        point = self.point + big_endian_to_int(h[:32])
        # chain code is the last 32 bytes
        chain_code = h[32:]
        # depth is current depth + 1
        depth = self.depth + 1
        # parent_fingerprint is the fingerprint of this node
        parent_fingerprint = self.fingerprint()
        # child number is the index
        child_number = index
        # return the HDPublicKey instance
        return HDPublicKey(
            point=point,
            chain_code=chain_code,
            depth=depth,
            parent_fingerprint=parent_fingerprint,
            child_number=child_number,
            network=self.network,
            pub_version=self.pub_version,
        )

    def traverse(self, path):
        """Returns the HDPublicKey at the path indicated.
        Path should be in the form of m/x/y/z."""

        if not path.startswith("m"):
            raise ValueError(f"Invalid Path: {path}")

        # accept path in uppercase and/or using h instead of '
        path = path.lower().replace("h", "'")

        # start current node at self
        current = self
        # get components of the path split at '/', ignoring the first
        components = path.split("/")[1:]
        # iterate through the components
        for child in components:
            # raise a ValueError if the path ends with a '
            if child[-1:] == "'":
                raise ValueError("HDPublicKey cannot get hardened child")
            # traverse the next child at the index
            current = current.child(int(child))
        # return the current node
        return current

    def raw_serialize(self):
        # start with pub_version, which should be a constant depending on network
        if self._raw is None:
            pub_version = XPUB[self.network]
            self._raw = self._serialize(pub_version)
        return self._raw

    def _serialize(self, pub_version):
        # start with the pub_version
        raw = pub_version
        # add the depth using int_to_byte
        raw += int_to_byte(self.depth)
        # add the parent_fingerprint
        raw += self.parent_fingerprint
        # add the child number in 4 bytes using int_to_big_endian
        raw += int_to_big_endian(self.child_number, 4)
        # add the chain code
        raw += self.chain_code
        # add the SEC pubkey
        raw += self.point.sec()
        return raw

    def xpub(self, version=None):
        """Returns the base58-encoded x/y/z pub."""
        if version is None:
            version = self.pub_version
        # get the serialization
        raw = self._serialize(version)
        # base58-encode the whole thing
        return encode_base58_checksum(raw)

    @classmethod
    def parse(cls, s):
        """Returns a HDPublicKey from an extended key string"""
        # get the bytes from the base58 using raw_decode_base58
        raw = raw_decode_base58(s)
        # check that the length of the raw is 78 bytes, otherwise raise ValueError
        if len(raw) != 78:
            raise ValueError("Not a proper extended key")
        # create a stream
        stream = BytesIO(raw)
        # return the raw parsing of the stream
        return cls.raw_parse(stream)

    @classmethod
    def raw_parse(cls, s):
        """Returns a HDPublicKey from a stream"""
        # first 4 bytes are the pub_version
        pub_version = s.read(4)
        # check that the pub_version is one of the TESTNET or MAINNET
        #  public keys, if not raise a ValueError
        if pub_version in ALL_TESTNET_XPUBS:
            network = "testnet"
        elif pub_version in ALL_MAINNET_XPUBS:
            network = "mainnet"
        else:
            raise ValueError("not a valid [t-z]pub pub_version: {pub_version}")
        # the next byte is depth
        depth = byte_to_int(s.read(1))
        # next 4 bytes are the parent_fingerprint
        parent_fingerprint = s.read(4)
        # next 4 bytes is the child number in big-endian
        child_number = big_endian_to_int(s.read(4))
        # next 32 bytes are the chain code
        chain_code = s.read(32)
        # last 33 bytes should be the SEC
        point = S256Point.parse(s.read(33))
        # return an instance of the class
        return cls(
            point=point,
            chain_code=chain_code,
            depth=depth,
            parent_fingerprint=parent_fingerprint,
            child_number=child_number,
            network=network,
            pub_version=pub_version,
        )


def calc_valid_seedpicker_checksums(first_words):
    """
    Generator to return all valid seedpicker checksums.

    For normal useage, just grab the first one only
    """
    for word in BIP39:
        try:
            HDPrivateKey.from_mnemonic(first_words + " " + word)
            yield (word)
        except InvalidChecksumWordsError:
            pass


def calc_num_valid_seedpicker_checksums(num_first_words):
    """
    If you have K firstwords, you will have V valid checksum words to choose from
    This convenience method is used for displaying progress to the user.
    """
    return {
        11: 128,
        14: 64,
        17: 32,
        20: 16,
        23: 8,
    }[num_first_words]


def is_valid_bip32_path(path):
    path = path.lower().strip().replace("'", "h").replace("//", "/")  # be forgiving

    if path == "m":
        # Not really a path, but since this is valid in the library (traverses to itself) we're counting it as valid
        return True

    if not path.startswith("m/"):
        return False

    sub_paths = path[2:].split("/")
    if len(sub_paths) >= 256:
        # https://bitcoin.stackexchange.com/a/92057
        return False

    for sub_path in sub_paths:
        if sub_path.endswith("h"):
            sub_path = sub_path[:-1]
        if not is_intable(sub_path):
            return False
        if int(sub_path) < 0:
            return False
        if int(sub_path) >= 2 ** 31:
            # https://bitcoin.stackexchange.com/a/92057
            return False

    return True


def ltrim_path(bip32_path, depth):
    """
    Left trim off a path by a given depth
    """
    if not is_valid_bip32_path(bip32_path):
        raise ValueError(f"Invalid bip32 path: {bip32_path}")

    # be forgiving
    path = bip32_path.lower().strip().replace("'", "h").replace("//", "/")

    if path.count("/") < depth:
        raise ValueError(
            f"Cannot left trim off a depth of {depth} from a path this short: {bip32_path}"
        )

    to_return = path.split("/")[depth + 1 :]
    return "m/" + "/".join(to_return)


def get_unhardened_child_path(base_path, root_path):
    """
    Return the difference between the base_path and root_path.

    Return None if there is no child_path (or it require hardened derivation).
    """
    if not is_valid_bip32_path(root_path):
        raise ValueError(f"Invalid bip32_path: {root_path}")

    base_path = base_path.strip().lower().replace("h", "'")
    root_path = root_path.strip().lower().replace("h", "'")

    if root_path.startswith(base_path):
        child_path = root_path[len(base_path) :]
        if "'" not in child_path:
            return f"m{child_path}"



from os import path
from secrets import randbits
from time import time

from buidl.helper import big_endian_to_int, int_to_big_endian, sha256


class InvalidBIP39Length(Exception):
    pass


class InvalidChecksumWordsError(Exception):
    pass


def secure_mnemonic(num_bits=256, extra_entropy=0):
    """
    Generates a mnemonic phrase using num_bits of entropy
    extra_entropy is optional and should not be saved as it is NOT SUFFICIENT to recover your mnemonic.
    extra_entropy exists only to prevent 100% reliance on your random number generator.
    """
    if num_bits not in (128, 160, 192, 224, 256):
        raise ValueError(f"Invalid num_bits: {num_bits}")
    if type(extra_entropy) is not int:
        raise TypeError(f"extra_entropy must be an int: {extra_entropy}")
    if extra_entropy < 0:
        raise ValueError(f"extra_entropy cannot be negative: {extra_entropy}")

    # if we have more bits than needed, mask so we get what we need
    if len(bin(extra_entropy)) > num_bits + 2:
        extra_entropy &= (1 << num_bits) - 1

    # For added paranoia, xor current epoch to extra_entropy
    # Would use time.time_ns() but that requires python3.7
    extra_entropy ^= int(time() * 1_000_000)

    # xor some random bits with the extra_entropy that was passed in
    preseed = randbits(num_bits) ^ extra_entropy
    # convert the number to big-endian
    s = int_to_big_endian(preseed, num_bits // 8)
    # convert to mnemonic
    mnemonic = bytes_to_mnemonic(s, num_bits)
    # sanity check
    if mnemonic_to_bytes(mnemonic) != s:
        raise RuntimeError("Generated mnemonic does not correspond to random bits")
    return mnemonic


def mnemonic_to_bytes(mnemonic):
    """returns a byte representation of the mnemonic"""
    all_bits = 0
    words = mnemonic.split()
    # check that there are 12, 15, 18, 21 or 24 words
    # if not, raise a ValueError
    if len(words) not in (12, 15, 18, 21, 24):
        raise InvalidBIP39Length(
            f"{len(words)} words (you need 12, 15, 18, 21, or 24 words)"
        )
    num_words = len(words)
    for word in words:
        all_bits <<= 11
        all_bits += BIP39[word]
    num_checksum_bits = num_words // 3
    checksum = all_bits & ((1 << num_checksum_bits) - 1)
    all_bits >>= num_checksum_bits
    num_bytes = (num_words * 11 - num_checksum_bits) // 8
    s = int_to_big_endian(all_bits, num_bytes)
    computed_checksum = sha256(s)[0] >> (8 - num_checksum_bits)
    if checksum != computed_checksum:
        raise InvalidChecksumWordsError("Checksum is wrong")
    return s


def bytes_to_mnemonic(b, num_bits):
    """returns a mnemonic given a byte representation"""
    if num_bits not in (128, 160, 192, 224, 256):
        raise InvalidBIP39Length(
            f"{num_bits} bits (you need 128, 160, 192, 224 or 256 bits)"
        )
    preseed = big_endian_to_int(b)
    # 1 extra bit for checksum is needed per 32 bits
    num_checksum_bits = num_bits // 32
    # the checksum is the sha256's first n bits. At most this is 8
    checksum = sha256(b)[0] >> (8 - num_checksum_bits)
    # we concatenate the checksum to the preseed
    all_bits = (preseed << num_checksum_bits) | checksum
    # now we get the mnemonic passphrase
    mnemonic = []
    # now group into groups of 11 bits
    for _ in range((num_bits + num_checksum_bits) // 11):
        # grab the last 11 bits
        current = all_bits & ((1 << 11) - 1)
        # insert the correct word at the front
        mnemonic.insert(0, BIP39[current])
        # shift by 11 bits so we can move to the next set
        all_bits >>= 11
    # return the mnemonic phrase by putting spaces between
    return " ".join(mnemonic)


class WordList:
    def __init__(self, filename, num_words):
        word_file = path.join(path.dirname(__file__), filename)
        with open(word_file, "r") as f:
            self.words = f.read().split()
        if len(self.words) != num_words:
            raise ValueError(f"Expected {num_words} but got {len(self.words)}")
        self.lookup = {}
        for i, word in enumerate(self.words):
            # add the word's index in the dict lookup
            self.lookup[word] = i
            # if the word is more than 4 characters, also keep
            #  a lookup of just the first 4 characters
            if len(word) > 4:
                self.lookup[word[:4]] = i

    def __getitem__(self, key):
        if type(key) == str:
            return self.lookup[key]
        elif type(key) == int:
            return self.words[key]
        else:
            raise KeyError("key needs to be a str or int")

    def __iter__(self):
        for word in self.words:
            yield word

    def __contains__(self, key):
        return key in self.words

    def normalize(self, word):
        return self[self[word.lower()]]


BIP39 = WordList("bip39_words.txt", 2048)
